import os
import logging
import logging.handlers
import abc
import inspect
import random
import string
import typing
import jinja2
import copy
import sys
import yaml
import jsonschema
from ansible.plugins.filter.core import FilterModule
from .exceptions import (AnsibleWorkflowConfigurationError, AnsibleWorkflowImportMissingBlock, AnsibleWorkflowRecursiveImport,
                         AnsibleWorkflowUnsupportedVersion, AnsibleWorkflowValidationError,
                         AnsibleWorkflowVaultScriptNotExists,
                         AnsibleWorkflowVaultScriptNotSet, AnsibleWorkflowYAMLNotValid)
from .engine import AnsibleWorkflow
from .models import Node, PNode, BNode, INode, CNode
from .validation import validate_workflow
from collections.abc import Mapping
from enum import Enum

class WorkflowLoader(metaclass=abc.ABCMeta):
    '''
    An abstract class that need to be implemented in order to support the
    parsing of a workflow file
    '''
    @abc.abstractmethod
    def parse(self, extra_vars: typing.Dict[str, str]) -> AnsibleWorkflow:
        pass


class YamlKeys(Enum):
    META_KEY: str = 'meta'
    DEFAULTS_KEY: str = 'defaults'
    OPTIONS_KEY: str = 'options'
    TEMPLATING_KEY: str = 'templating'


class WorkflowYamlLoader(WorkflowLoader):
    '''
    Handles loading of a workflow from a yml file with directives that mimics
    Ansible task names.
    '''
    def __init__(self, workflow_file: str, logging_dir: str,
                 logging_level: str = None, input_templating: dict={},
                 check_mode=False, verbosity=0, doubtful_mode=False):
        '''
        Initialize the loader
        Args:
            workflow (string): The relative or absolute path to the file to
                be loaded.
            logger (logging.Logger): A logging.logger instance
        Raises:
            AnsibleWorkflowYAMLNotValid: If the file cannot be loaded or is not yaml
        Returns:
            WorkflowLoader: An instance of the abstract class WorkflowLoader
        '''
        self.__define_logger(logging_dir, logging_level)
        self.__workflow_file: string = workflow_file
        self.__workflow: AnsibleWorkflow = AnsibleWorkflow(self.__workflow_file, logging_dir, logging_level, doubtful_mode)
        self._default_format_version: int = 1
        self.__check_mode: bool = check_mode
        self.__verbosity = verbosity
        self.input_templating: dict = input_templating

        # initialize template environment
        self._template_env = jinja2.Environment(loader=jinja2.FileSystemLoader("."), undefined=jinja2.StrictUndefined)
        self._template_env.filters.update(FilterModule().filters())
        # contains the dictionary parsed from the yaml workflow file
        try:
            self.__yaml_parsed: dict = self._load_yaml(self.get_contents(self.__workflow_file))
        except yaml.scanner.ScannerError as yerr:
            self._logger.exception("Errors in YAML file format. %s" % yerr)
            raise AnsibleWorkflowYAMLNotValid(yerr.message)
        except Exception as err:
            self._logger.error("Impossible to load workflow file. %s" % err)
            raise AnsibleWorkflowYAMLNotValid(err)

    def __define_logger(self, logging_dir: str, logging_level: str):
        '''
        Define a logger specific for this class
        Args:
            logging_dir (string): The directory path where log
            logging_level (string): The level wich the logger should be set
        '''
        logger_name = self.__class__.__name__
        logger_file_path = os.path.join(logging_dir, 'loader.log')
        if not os.path.exists(os.path.dirname(logger_file_path)):
            os.makedirs(os.path.dirname(logger_file_path))

        logger = logging.getLogger(logger_name)
        if logging_level:
            logger.setLevel(getattr(logging, logging_level.upper()))
        logger_handler = logging.handlers.TimedRotatingFileHandler(
            logger_file_path,
            when='d',
            backupCount=3,
            encoding='utf8'
        )
        logger_handler.setFormatter(logging.Formatter('%(asctime)s [%(levelname)s] %(name)s: %(message)s'))
        logger.addHandler(logger_handler)
        self._logger = logger
        self._logging_dir = logging_dir

    def _load_yaml(self, contents: str):
        '''
        Attempts to deserialize the contents of a YAML object
        Args:
            contents (string): The contents to deserialize
        Returns:
            dict: If the contents are YAML serialized
            None: if the contents are not YAML serialized
        '''
        return yaml.safe_load(contents)

    def _write_yaml(self, contents: dict, path: str):
        '''
        Attempts to serialize an object to a YAML string into a file
        Args:
            contents (string): The contents to write
            path (string): The path of the file
        '''
        with open(path, 'w') as file:
            yaml.dump(contents, file)

    def get_contents(self, path: str) -> str:
        '''
        Loads the contents of the file specified by path
        Args:
            path (string): The relative or absolute path to the file to
                be loaded.  If the path is relative, then it is combined
                with the base_path to generate a full path string
        Returns:
            string: The contents of the file as a string
        Raises:
            AnsibleWorkflowConfigurationError: If the file cannot be loaded
        '''
        try:
            if not os.path.exists(path):
                raise AnsibleWorkflowConfigurationError('Specified workflow path does not exist: %s' % path)

            data = ''

            prependwf_file = '{}/_wf.yml'.format(os.path.dirname(os.path.realpath(path)))

            if os.path.exists(prependwf_file):
                with open(prependwf_file) as f:
                    data += f.read()

            with open(path) as f:
                data += f.read()

            return data

        except (IOError, OSError) as err:
            self._logger.exception(err)
            raise AnsibleWorkflowConfigurationError('Error trying to load workflow file contents: %s' % err)

    def parse(self, extra_vars: typing.Dict[str, str]) -> AnsibleWorkflow:
        '''
        Parse the workflow file, checking the correctness of the format depending by the version of the format
        Args:
            extra_vars (dict): A dictionary containing the extra variables to be passed to playbooks
        Returns:
            AnsibleWorkflow: An AnsibleWorkflow class instance
        Raises:
            ValidationError: If the input file doesn't match the schema
            AnsibleWorkflowUnsupportedVersion: If the version format is not supported
        '''
        # grab the meta key to verify the version specified
        if isinstance(self.__yaml_parsed, Mapping):
            yml_format_version = self.__yaml_parsed.get(YamlKeys.META_KEY.value, {}).get('format-version', '1')
        else:
            yml_format_version = '1'

        # search a method that is able to parse this versione
        _method_name: str = '_parse_v' + str(yml_format_version)
        _method_func = getattr(self, _method_name)
        if _method_func is None:
            raise AnsibleWorkflowUnsupportedVersion

        # call the parser
        _method_func(extra_vars)
        # return the workflow instance
        return self.__workflow

    def _parse_v1(self, extra_vars: typing.Dict[str, str]):
        '''
        Parse the yaml version 1 workflow file
        Args:
            extra_vars (dict): A dictionary containing the extra variables to be passed to playbooks
        Raises:
            AnsibleWorkflowVaultScriptNotExists: If the workflow file doesn't exists
        '''
        # validate the workflow file accordingly with the format version
        schema_path = os.path.join(os.path.dirname(__file__), '..', 'schemas', 'v1.json')
        try:
            validate_workflow(self.__yaml_parsed, schema_path)
        except jsonschema.ValidationError as err:
            self._logger.fatal("Impossible to parse the workflow. Validation error: %s" % err)
            raise AnsibleWorkflowValidationError("%s" % err.message)

        options: dict = dict()
        defaults: dict = dict()
        workflow: dict = dict()
        template_variables: dict = dict()

        # get values from options and default keys of the file
        defaults['inventory'] = self.__yaml_parsed.get(YamlKeys.DEFAULTS_KEY.value, {}).get('inventory')
        defaults['verbosity'] = self.__verbosity if self.__verbosity != 0 else self.__yaml_parsed.get(YamlKeys.DEFAULTS_KEY.value, {}).get('verbosity', 0)
        defaults['vars'] = extra_vars if extra_vars else self.__yaml_parsed.get(YamlKeys.DEFAULTS_KEY.value, {}).get('vars', {})
        defaults['vault_ids'] = self.__yaml_parsed.get(YamlKeys.DEFAULTS_KEY.value, {}).get('vault_ids', [])
        defaults['project_path'] = self.__yaml_parsed.get(YamlKeys.DEFAULTS_KEY.value, {}).get('project_path', '.')
        defaults['limit'] = self.__yaml_parsed.get(YamlKeys.DEFAULTS_KEY.value, {}).get('limit')
        options['vault_script'] = self.__yaml_parsed.get(YamlKeys.OPTIONS_KEY.value, {}).get('vault_script')
        options['global_path'] = self.__yaml_parsed.get(YamlKeys.OPTIONS_KEY.value, {}).get('global_path')

        workflow_dir = os.path.dirname(os.path.abspath(self.__workflow_file))
        if options.get('global_path', False):
            if not os.path.isabs(options['global_path']):
                options['global_path'] = os.path.join(workflow_dir, options['global_path'])
        else:
            options["global_path"] = workflow_dir


        template_variables = {
            **self.input_templating,
            **self.__yaml_parsed.get(YamlKeys.TEMPLATING_KEY.value, {})
        }

        self._perform_template_rendering(template_variables, self.input_templating)
        self._perform_template_rendering(defaults, template_variables)
        self._perform_template_rendering(options, template_variables)

        # assign calculated string to be rendered
        self.__yaml_parsed[YamlKeys.TEMPLATING_KEY.value] = template_variables
        self.__yaml_parsed[YamlKeys.OPTIONS_KEY.value] = options
        self.__yaml_parsed[YamlKeys.DEFAULTS_KEY.value] = defaults

        # prepend global path to vault script
        if options.get("vault_script", False):
            if options.get("global_path", False):
                if options.get("vault_script", False) and not os.path.isabs(options["vault_script"]):
                    options["vault_script"] = os.path.join(options["global_path"], options["vault_script"])

            # calculate the absolute path for the vault script
            options["vault_script"] = os.path.abspath(options["vault_script"])
            if not os.path.exists(options["vault_script"]):
                self._logger.error("Vault script %s doesn't exists " % (options["vault_script"]))
                raise AnsibleWorkflowVaultScriptNotExists("Vault script %s doesn't exists " % (options["vault_script"]))

        # log this volues for debugging purpose
        self._logger.info("---------- [defaults] ----------")
        for default_key in defaults.keys():
            self._logger.info("%s : %s" % (default_key, defaults[default_key]))
        self._logger.info("---------- [options] ----------")
        for option_key in options.keys():
            self._logger.info("%s : %s" % (option_key, options[option_key]))
        self._logger.info("---------- [template variables] ----------")
        for template_key in template_variables.keys():
            self._logger.info("%s : %s" % (template_key, template_variables[template_key]))
        self._logger.info("-------------------------------")

        # call the creation of the workflow adding a start and end block node
        self.__yaml_parsed['workflow'].insert(0, dict(id='_s', block=[]))
        self.__yaml_parsed['workflow'].append(dict(id='_e', block=[]))

        # add a dummy root node to visualize the workflow like a tree in console
        self.__workflow.add_node(BNode("_root"))
        self.__workflow.get_original_graph().add_node('_root')

        # perform static inclusion where include_block is found
        self._perform_static_inclusion(self.__yaml_parsed['workflow'], options, template_variables=template_variables)

        # perform templating of strings
        # self._perform_template_rendering(self.__yaml_parsed['workflow'], template_variables)
        #self._perform_template_rendering(self.__yaml_parsed, template_variables)
        self._write_yaml(self.__yaml_parsed, os.path.join(self._logging_dir, 'rendered_workflow.yml'))

        # validate twice after the file inclusion
        try:
            validate_workflow(self.__yaml_parsed, schema_path)
        except jsonschema.ValidationError as err:
            self._logger.fatal("Impossible to parse the workflow. Validation error: %s" % err)
            raise AnsibleWorkflowValidationError("Wrong workflow format.\n%s" % err.message)

        # call the parser of the workflow key
        self._parse_workflow_v1(to_be_imported=self.__yaml_parsed['workflow'], parent_nodes=[], strategy='serial', defaults=defaults, options=options)

    def _perform_string_template_rendering(self, template_string: str, template_variables: typing.Dict[str, typing.Any]):
        try:
            template = self._template_env.from_string(template_string)
            return template.render(**template_variables)
        except jinja2.exceptions.UndefinedError as jerr:
            self._logger.error(f"Templating error: {jerr}. Rendered block: {template_string}")
            raise jerr

    def _perform_template_rendering(self, parsed_yml: typing.Any, template_variables: typing.Dict[str, typing.Any]):
        '''
        Search inside the workflow values and apply jinja templating
        Args:
            parsed_yml (list): A list of dict containing workflow nodes in a block
            template_variables (dict): The variables to be applied to the template
        '''
        if isinstance(parsed_yml, dict):
            for key, value in parsed_yml.items():
                if isinstance(value, str):
                    parsed_yml[key] = self._perform_string_template_rendering(value, template_variables)
                    self._logger.debug("Templating: %s vars: %s" % (parsed_yml[key], template_variables))
                elif isinstance(value, dict) or isinstance(value, list):
                    self._perform_template_rendering(value, template_variables)
        elif isinstance(parsed_yml, list):
            i = 0
            for value in parsed_yml:
                if isinstance(value, str):
                    parsed_yml[i] =  self._perform_string_template_rendering(value, template_variables)
                    self._logger.debug("Templating: %s" % parsed_yml[i])
                elif isinstance(value, dict) or isinstance(value, list):
                    self._perform_template_rendering(value, template_variables)
                i = i + 1


    def _perform_static_inclusion(self, parsed_yml: typing.List[dict], options: typing.Dict[str, str], prefix='', template_variables: typing.Dict[str, typing.Any]={}):
        '''
        Perform inclusion of file containig block specifications
        Args:
            parsed_yml (list): A list of dict containing workflow nodes in a block
            defaults (dict): A dictionary with defaults settings
            prefix (str): The prefix to be added to the ids of the imported file
        Raises:
            AnsibleWorkflowImportMissingBlock: If the file imported do not contains a block directive
            AnsibleWorkflowRecursiveImport: If the imported file import the same file
        '''
        for inode in parsed_yml:
            current_template_variables = copy.copy(template_variables)
            if 'include_block' in inode:
                # perform templating for current inclusion
                temporary_template_variables = copy.copy(template_variables)
                if 'templating' in inode:
                    self._perform_template_rendering(inode['templating'], template_variables=current_template_variables)
                    temporary_template_variables.update(inode['templating'])

                # prepend included block file with global path if not absolute
                self._logger.debug("XXXX")
                to_be_included_file = self._perform_string_template_rendering(inode.pop('include_block'), temporary_template_variables)
                included_block_prefix = inode.pop('id_prefix', '')
                if options.get("global_path", False) and not os.path.isabs(to_be_included_file):
                    to_be_included_file = os.path.join(options["global_path"], to_be_included_file)
                block_file_parsed = self._load_yaml(self.get_contents(to_be_included_file))

                # copy the identifier of the importing node
                inode_id = copy.deepcopy(inode.get('id'))
                inode_templating = copy.deepcopy(inode.get('templating'))
                # update node keys
                inode.update(block_file_parsed)

                self._logger.info("Included block file %s into the workflow" % to_be_included_file)
                # set the importing id to the imported node
                if inode_id:
                    self._logger.info("Overwrited imported block id with the importing node id (%s)" % inode_id)
                    inode["id"] = inode_id

                if 'templating' in inode:
                    current_template_variables.update(inode['templating'])

                if inode_templating:
                    self._logger.info("Overwrited imported block templating (%s) with the importing node templating (%s)" % (inode['templating'], inode_templating))
                    inode["templating"] = inode_templating
                    current_template_variables.update(inode['templating'])
                    self._logger.debug("Resulting template variables: %s" % current_template_variables)

                if 'block' not in inode:
                    self._logger.error("The imported block file %s doesn't contain a block" % to_be_included_file)
                    raise AnsibleWorkflowImportMissingBlock("The imported block file %s doesn't contain a block" % to_be_included_file)
                # check if there is an inclusion of itself
                for subnode in inode['block']:
                    if 'include_block' in subnode and subnode['include_block'] in to_be_included_file:
                        self._logger.error("The imported block file %s is importing itself" % to_be_included_file)
                        raise AnsibleWorkflowRecursiveImport("The imported block file %s is importing itself" % to_be_included_file)

                self._perform_static_inclusion(inode['block'], options, included_block_prefix, template_variables=current_template_variables)
            elif 'block' in inode:
                inode['id'] = str(prefix) + str(inode['id'])
                if 'templating' in inode:
                    current_template_variables.update(inode['templating'])
                self._perform_static_inclusion(inode['block'], options, prefix, template_variables=current_template_variables)
                # perform also a render of the remaining field of a block after the leaf node are rendered
                self._perform_template_rendering(inode, template_variables=current_template_variables)
            else:
                inode['id'] = str(prefix) + str(inode['id'])
                self._logger.info("Templating block: %s - with variables: %s" % (inode['id'], current_template_variables))
                # perform the render on the lead of a tree
                self._perform_template_rendering(inode, template_variables=current_template_variables)

    def _parse_workflow_v1(self, to_be_imported: typing.List[dict], parent_nodes: typing.List[Node],
                           strategy: str, defaults: typing.Dict[str, str],
                           options: typing.Dict[str, str], level: int = 1, block_id: str = '_root'):
        '''
        Parse the workflow key of the file, adding nodes and links to the execution graph
        Args:
            to_be_imported (list): A list of dict containing workflow nodes in a block
            parent_nodes (list): A list of Nodes instances that are parents of the actual block
            strategy (str): The strategy of the block (can be serial or parallel)
            defaults (dict): A dictionary with defaults settings
            options(dict): A dictionary with global settings
            level (int): The nesting level of the block
            block_id (str): The parent block identifier
        Raises:
            AnsibleWorkflowVaultScriptNotSet: If a node specify some vault ids but the vault script is not set
        '''
        indentation = '\t' * (len(inspect.stack(0)) - 7)

        # init to loop over the structure
        zero_outdegree_nodes = []
        for inode in to_be_imported:
            # generate a node identifier and set to the node
            gnode_id = inode.get('id', ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(5)))
            gnode_id = str(gnode_id)
            inode['id'] = gnode_id

            self._logger.debug("-->> %s node: %s       parents: %s       zero_outdegree: %s" %
                               (indentation, inode['id'], [p.get_id() for p in parent_nodes], [p.get_id() for p in zero_outdegree_nodes]))

            for parent_node in parent_nodes:
                self._logger.debug("---- %s added graph link (%s) --> (%s)" % (indentation, parent_node.get_id(), gnode_id))
                self.__workflow.add_link(parent_node.get_id(), gnode_id)

            if strategy == 'serial':
                parent_nodes = []
                for zero_outdegree_node in zero_outdegree_nodes:
                    self.__workflow.add_link(zero_outdegree_node.get_id(), gnode_id)
                zero_outdegree_nodes = []

            # generate the object representing the graph
            if 'block' in inode:
                gnode = BNode(gnode_id)
                block_sub_nodes = self._parse_workflow_v1(inode['block'], [gnode, ], inode.get('strategy', 'parallel'), defaults, options, level + 1, gnode_id)
            else:
                if 'import_playbook' in inode:
                    # set the playbook
                    playbook = inode['import_playbook']
                    # init the parameters input
                    pnode_parameters = dict(id=gnode_id, playbook=playbook, artifact_dir=self._logging_dir, check_mode=self.__check_mode)

                    # set all the inode parameters or if not set defaults parameters
                    for parameter, default_or_inode_key in dict(inventory="inventory",
                                                                extra_vars="vars",
                                                                vault_ids="vault_ids",
                                                                project_path="project_path",
                                                                limit="limit",
                                                                verbosity="verbosity",
                                                                description="description",
                                                                reference="reference").items():
                        if inode.get(default_or_inode_key):
                            pnode_parameters[parameter] = inode[default_or_inode_key]
                        elif defaults.get(default_or_inode_key):
                            pnode_parameters[parameter] = defaults[default_or_inode_key]

                    # prepend global path to project and inventory
                    if options.get("global_path", False):
                        if not os.path.isabs(pnode_parameters["project_path"]):
                            pnode_parameters["project_path"] = os.path.join(options["global_path"], pnode_parameters["project_path"])
                        if not os.path.isabs(pnode_parameters["inventory"]):
                            pnode_parameters["inventory"] = os.path.join(options["global_path"], pnode_parameters["inventory"])

                    # add the vault script to all vault id
                    vault_ids = []

                    if len(pnode_parameters.get("vault_ids", [])) > 0 and options['vault_script'] is None:
                        raise AnsibleWorkflowVaultScriptNotSet("Vault script not set but vault IDs specified for node %s" % (gnode_id))

                    for vault_id in pnode_parameters.get("vault_ids", []):
                        vault_ids.append("%s@%s" % (vault_id, options["vault_script"]))
                    pnode_parameters["vault_ids"] = vault_ids

                    gnode = PNode(**pnode_parameters)
                    gnode.set_logger(self._logger)
                    self._logger.debug("---- %s added node: %s, level: %s, block type: %s, block id: %s" %
                                    (indentation, pnode_parameters, level, strategy, block_id))
                elif inode.get('checkpoint', False):
                    gnode = CNode(gnode_id, description=inode.get('description', ''), reference=inode.get('reference', ''))
                else: # It's a info node
                    gnode = INode(gnode_id, description=inode.get('description', ''), reference=inode.get('reference', ''))

            # the node specification is added
            self.__workflow.add_node(gnode, dict(level=level, block=dict(strategy=strategy, block_id=block_id)))

            # saves also original tree
            self.__workflow.get_original_graph().add_node(gnode_id)
            self.__workflow.get_original_graph().add_edge(block_id, gnode_id)

            if 'block' in inode:
                zero_outdegree_nodes.extend(block_sub_nodes)
            else: # This covers both playbook and dummy nodes
                if strategy == 'parallel' or (strategy == 'serial' and inode == to_be_imported[-1]):
                    zero_outdegree_nodes.append(gnode)

            # if the strategy is serial
            if strategy == 'serial':
                if 'block' in inode and len(inode['block']) > 0:
                    # add the node from the subtree as parent
                    parent_nodes = block_sub_nodes
                else:
                    # or add current node as the parent
                    parent_nodes = [gnode, ]
            self._logger.debug("<<-- %s node: %s       parents: %s       zero_outdegree: %s" %
                               (indentation, gnode_id, [p.get_id() for p in parent_nodes], [p.get_id() for p in zero_outdegree_nodes]))
        return zero_outdegree_nodes



import httpx
import logging
from typing import List, Dict, Any, Optional

class ApiClient:
    def __init__(self, base_url: str, logger=None):
        self.base_url = base_url
        self.client = httpx.Client(base_url=self.base_url)
        self.logger = logger or logging.getLogger(__name__)

    def get_workflow_status(self) -> Optional[Dict[str, Any]]:
        try:
            response = self.client.get("/workflow")
            response.raise_for_status()
            return response.json()
        except (httpx.ConnectError, httpx.HTTPStatusError):
            return None

    def get_all_nodes(self) -> Optional[List[Dict[str, Any]]]:
        try:
            response = self.client.get("/workflow/nodes")
            response.raise_for_status()
            return response.json()
        except (httpx.ConnectError, httpx.HTTPStatusError):
            return None

    def get_workflow_graph(self) -> Optional[List[List[str]]]:
        try:
            response = self.client.get("/workflow/graph")
            response.raise_for_status()
            return response.json()["edges"]
        except (httpx.ConnectError, httpx.HTTPStatusError):
            return None

    def get_node_stdout(self, node_id: str) -> Optional[str]:
        try:
            response = self.client.get(f"/workflow/node/{node_id}/stdout")
            response.raise_for_status()
            return response.json()["stdout"]
        except (httpx.ConnectError, httpx.HTTPStatusError):
            return None

    def stop_workflow(self, mode: str = "graceful"):
        try:
            response = self.client.post("/workflow/stop", json={"mode": mode})
            response.raise_for_status()
        except (httpx.ConnectError, httpx.HTTPStatusError):
            pass

    def pause_workflow(self):
        self.logger.info("Pausing workflow via API")
        try:
            response = self.client.post("/workflow/pause")
            response.raise_for_status()
        except (httpx.ConnectError, httpx.HTTPStatusError) as e:
            self.logger.error(f"Failed to pause workflow: {e}")
            pass

    def resume_workflow(self):
        self.logger.info("Resuming workflow via API")
        try:
            response = self.client.post("/workflow/resume")
            response.raise_for_status()
        except (httpx.ConnectError, httpx.HTTPStatusError) as e:
            self.logger.error(f"Failed to resume workflow: {e}")
            pass

    def shutdown_backend(self):
        try:
            self.client.post("/shutdown")
        except (httpx.ReadError, httpx.ConnectError, httpx.HTTPStatusError):
            # This is expected as the server will shut down before sending a response
            pass

    def restart_node(self, node_id: str):
        try:
            response = self.client.post(f"/workflow/node/{node_id}/restart")
            response.raise_for_status()
        except (httpx.ConnectError, httpx.HTTPStatusError):
            pass

    def skip_node(self, node_id: str):
        try:
            response = self.client.post(f"/workflow/node/{node_id}/skip")
            response.raise_for_status()
        except (httpx.ConnectError, httpx.HTTPStatusError):
            pass

    def approve_node(self, node_id: str):
        try:
            response = self.client.post(f"/workflow/node/{node_id}/approve")
            response.raise_for_status()
        except (httpx.ConnectError, httpx.HTTPStatusError):
            pass

    def disapprove_node(self, node_id: str):
        try:
            response = self.client.post(f"/workflow/node/{node_id}/disapprove")
            response.raise_for_status()
        except (httpx.ConnectError, httpx.HTTPStatusError):
            pass

    def check_health(self) -> bool:
        try:
            response = self.client.get("/health")
            response.raise_for_status()
            return True
        except (httpx.ConnectError, httpx.HTTPStatusError):
            return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2021 BartokIT
# Ansible® is a registered trademark of Red Hat, Inc. in the United States and other countries.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License v3.0 as published by
# the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the  GNU Affero General Public License v3.0
# along with this program; if not, see <https://www.gnu.org/licenses/agpl.html/>.

import logging
import logging.handlers
import sys
import argparse
import os
import os.path
from datetime import datetime
import threading
import signal
from rich.console import Console
import httpx
import subprocess
import time

from .ui.stdout import StdoutWorkflowOutput
from .ui.textual import TextualWorkflowOutput
from ansible.cli.arguments import option_helpers as opt_help
from ansible.parsing.splitter import parse_kv

BACKEND_URL = "http://127.0.0.1:8001"

def define_logger(logging_dir, level):
    logger_file_path = os.path.join(logging_dir, 'main.log')
    if not os.path.exists(os.path.dirname(logger_file_path)):
        os.makedirs(os.path.dirname(logger_file_path))

    logger = logging.getLogger('main')
    if level:
        logger.setLevel(getattr(logging, level.upper()))
    logger_handler = logging.handlers.TimedRotatingFileHandler(
        logger_file_path,
        when='d',
        backupCount=3,
        encoding='utf8'
    )
    logger_handler.setFormatter(logging.Formatter('%(asctime)s [%(levelname)s] %(name)s: %(message)s'))
    logger.addHandler(logger_handler)
    return logger


def check_and_start_backend(logger):
    try:
        httpx.get(f"{BACKEND_URL}/health")
        logger.info("Backend is already running.")
    except httpx.ConnectError:
        logger.info("Backend not running. Starting it now.")
        project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))

        popen_kwargs = {
            "cwd": project_root,
            "stdout": open(os.path.join(project_root, "backend_stdout.log"), "wb"),
            "stderr": open(os.path.join(project_root, "backend_stderr.log"), "wb"),
        }

        if os.name == 'nt':
            popen_kwargs['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP
        elif os.name == 'posix':
            popen_kwargs['start_new_session'] = True

        process = subprocess.Popen(
            [sys.executable, "-m", "uvicorn", "ansible_plan.service:app", "--port", "8001"],
            **popen_kwargs
        )

        for _ in range(10):
            try:
                httpx.get(f"{BACKEND_URL}/health")
                logger.info("Backend started successfully.")
                return process
            except httpx.ConnectError:
                time.sleep(1)
        logger.error("Failed to start the backend.")
        sys.exit(1)
    return None


def read_options():
    parser = argparse.ArgumentParser(description='This programs mimics the AWX/Ansible Tower® workflows from command line.')
    parser.add_argument('workflow', type=str, help='Workflow file')

    # influences workflow execution
    group = parser.add_mutually_exclusive_group()

    group.add_argument('--execute-nodes', dest='filter_nodes', default="",
                        help='filter nodes to be executed inside a workflow, executing only the selected (comma separated).')

    group.add_argument('--skip-nodes', dest='skip_nodes', default="",
                        help='Filter nodes to be executed inside a workflow, skipping the selected (comma separated)')

    parser.add_argument('-sn', '--start-from-node', dest='start_from_node', default="",
                        help='Start the execution of the workflow from the specified node')

    parser.add_argument('-en', '--end-to-node', dest='end_to_node', default="",
                        help='Stop the execution of the workflow to the specified node')

    parser.add_argument('-vo', '--verify-only', dest='verify_only', action='store_true',
                        help='Perform only the verification of the workflow content')

    parser.add_argument('-c', '--check', dest='check_mode', action='store_true',
                        help='Launch all playbook in check mode')

    parser.add_argument('-v', dest='verbosity', action='count', default=0,
                        help='The logging level of the playbook')

    parser.add_argument('-nir', '--no-interactive-retry', dest='interactive_retry', action='store_false',
                        help='Avoid interactive retry of job')

    parser.add_argument('--doubtful-mode', dest='doubtful_mode', action='store_true',
                        help='Ask for each node if should be started or skipped')

    # add extra vars parameter from ansible library
    opt_help.add_runtask_options(parser)

    # influences the output
    parser.add_argument('--mode', default='stdout', choices=["stdout", "textual"],
                        help='Render the progress using textual or stdout.')

    parser.add_argument('-d', '--draw', dest='draw_png', action='store_true',
                        help='Output also a PNG of the graph inside the log folder')

    parser.add_argument('-dpi', '--draw-dpi', dest='draw_dpi', default=72, type=int,
                        help='Choose the dpi of the draw graph')

    parser.add_argument('-size', '--draw-size', dest='draw_size', default=10, type=int,
                        help='Choose the size of the draw graph')

    parser.add_argument('--log-dir', dest='log_dir', default='logs',
                        help='set the parent output logging directory. defaults to logs/[workflow name]-[execution time]')

    parser.add_argument('--log-dir-no-info', dest='log_dir_no_info', action='store_true',
                        help='Does not add the workflow name and the execution time to the log dir name')

    parser.add_argument('--log-level', dest='log_level', default='info', choices=["debug", "info", "warning", "error", "critical"],
                        help='set the logging level. defaults to info')

    # loader input
    parser.add_argument('-it', '--input-templating', dest='input_templating', default=[], action='append', type=keyvalue,
                        help='Input variable for the templating -it key1=value1 -it key2=value2')

    return parser.parse_args()

def keyvalue(value):
    if '=' not in value:
        raise Exception('Key value malformatted: key=value, missing the "="')
    return value.split('=')

def main():
    os.environ['TERM'] = 'xterm-256color'
    cmd_args = read_options()

    logging_dir = "%s" % cmd_args.log_dir
    if not cmd_args.log_dir_no_info:
        logging_dir += "/%s_%s" % (os.path.basename(cmd_args.workflow), datetime.now().strftime("%Y%m%d_%H%M%S"))

    logger = define_logger(logging_dir, cmd_args.log_level)
    console = Console()

    check_and_start_backend(logger)

    extra_vars = {}
    for single_extra_vars in cmd_args.extra_vars:
        extra_vars.update(parse_kv(single_extra_vars))

    input_templating = {x: y for [x, y] in cmd_args.input_templating}

    start_payload = {
        "workflow_file": os.path.abspath(cmd_args.workflow),
        "extra_vars": extra_vars,
        "input_templating": input_templating,
        "check_mode": cmd_args.check_mode,
        "verbosity": cmd_args.verbosity,
        "start_from_node": cmd_args.start_from_node,
        "end_to_node": cmd_args.end_to_node,
        "skip_nodes": cmd_args.skip_nodes.split(",") if cmd_args.skip_nodes else [],
        "filter_nodes": cmd_args.filter_nodes.split(",") if cmd_args.filter_nodes else [],
        "log_dir": cmd_args.log_dir,
        "log_dir_no_info": cmd_args.log_dir_no_info,
        "log_level": cmd_args.log_level,
        "verify_only": cmd_args.verify_only,
        "doubtful_mode": cmd_args.doubtful_mode,
    }

    try:
        response = httpx.post(f"{BACKEND_URL}/workflow", json=start_payload, timeout=30)
        response.raise_for_status()
        response_data = response.json()
        if response_data.get("status") == "reconnected":
            console.print(f"Reconnected to running workflow: {cmd_args.workflow}")

    except httpx.HTTPStatusError as e:
        if e.response.status_code == 409:
            detail = e.response.json().get("detail", {})
            message = detail.get("message", "An unknown conflict occurred.")
            running_workflow_path = detail.get("running_workflow_file", "")

            console.print(f"[bold red]Error:[/bold red] {message}")

            if running_workflow_path:
                y_or_n = console.input(f"Do you want to connect to the running workflow at '{running_workflow_path}'? [y/n] ")
                if y_or_n.lower() == 'y':
                    cmd_args.workflow = running_workflow_path
                    console.print(f"Connecting to existing workflow: {cmd_args.workflow}")
                else:
                    sys.exit(0)
            else:
                sys.exit(1)
        else:
            logger.error(f"Failed to start workflow: {e}")
            print(f"Failed to start workflow: {e}", file=sys.stderr)
            if hasattr(e, 'response') and e.response:
                print(e.response.text, file=sys.stderr)
            sys.exit(1)
    except httpx.ConnectError as e:
        logger.error(f"Failed to start workflow: {e}")
        print(f"Failed to start workflow: {e}", file=sys.stderr)
        sys.exit(1)


    if cmd_args.mode == 'textual':
        output = TextualWorkflowOutput(
            backend_url=BACKEND_URL,
            event=threading.Event(),
            logging_dir=logging_dir,
            log_level=cmd_args.log_level,
            cmd_args=cmd_args
        )
        output.run()

        try:
            response = httpx.get(f"{BACKEND_URL}/workflow")
            response.raise_for_status()
            status = response.json().get("status")
            if status == "running":
                console.print("\nDetaching from workflow. The backend will continue to run.")
                console.print("To re-attach, run the same command again.")
        except (httpx.ConnectError, httpx.HTTPStatusError):
            pass
    else:
        stdout_thread = StdoutWorkflowOutput(
            backend_url=BACKEND_URL,
            event=threading.Event(),
            logging_dir=logging_dir,
            log_level=cmd_args.log_level,
            cmd_args=cmd_args
        )
        stdout_thread.start()

        def signal_handler(sig, frame):
            console.print("\nDetaching from workflow. The backend will continue to run.")
            console.print("To re-attach, run the same command again.")
            stdout_thread.event.set()
            # wait for the thread to finish
            stdout_thread.join()
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)
        stdout_thread.join()

    # Shutdown logic
    try:
        response = httpx.get(f"{BACKEND_URL}/workflow")
        response.raise_for_status()
        status = response.json().get("status")
        if status != "running":
            logger.info("Workflow finished. Shutting down backend.")
            httpx.post(f"{BACKEND_URL}/shutdown")
    except (httpx.ConnectError, httpx.HTTPStatusError) as e:
        logger.warning(f"Could not get workflow status or shutdown backend: {e}")


if __name__ == "__main__":
    main()



import typing
import warnings
import threading
from datetime import datetime
import time
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", message="networkx backend defined more than once: nx-loopback")
    import networkx as nx
import os
import logging
import logging.handlers
from .exceptions import AnsibleWorkflowDuplicateNodeId, AnsibleWorkflowPlaybookNodeCheck
from .models import WorkflowStatus, NodeStatus, Node, BNode, PNode, CNode, INode, WorkflowEventType, WorkflowEvent, WorkflowListener


class AnsibleWorkflow():
    def __init__(self, workflow_file, logging_dir, log_level, doubtful_mode: bool = False, filtered_nodes=None):
        self.__graph: nx.DiGraph = nx.DiGraph()
        self.__original_graph: nx.DiGraph = nx.DiGraph()
        self.__running_status = WorkflowStatus.NOT_STARTED
        self.__define_logger(logging_dir, log_level)
        self.__data = dict()
        self.__running_nodes = []
        self.__stopped = False
        self.__listeners: WorkflowListener = []
        self.__skipped_nodes: typing.List[str] = []
        self.__logging_dir = logging_dir
        self.__workflow_file = workflow_file
        self.__resume_event = threading.Event()
        self._validation_errors = []
        self.__pause_event = threading.Event()
        self.__pause_event.set()
        self.__stopping = False
        self.__doubtful_mode = doubtful_mode

    def get_validation_errors(self):
        return self._validation_errors

    def add_validation_error(self, error: str):
        self._validation_errors.append(error)

    def set_status(self, status: WorkflowStatus):
        self.__running_status = status

    def get_workflow_file(self):
        return self.__workflow_file

    def get_logging_dir(self):
        return self.__logging_dir

    def set_filtered_nodes(self, filter_nodes: typing.List[str]):
        if len(filter_nodes) > 0:
            remaining_nodes = set(self.__graph.nodes) - set(filter_nodes)
            self.__skipped_nodes = remaining_nodes

    def set_skipped_nodes(self, skipped_nodes: typing.List[str]):
        if len(skipped_nodes) > 0:
            self.__skipped_nodes = skipped_nodes

    def __define_logger(self, logging_dir, level):
        logger_name = self.__class__.__name__
        logger_file_path = os.path.join(logging_dir, 'workflow.log')
        if not os.path.exists(os.path.dirname(logger_file_path)):
            os.makedirs(os.path.dirname(logger_file_path))

        logger = logging.getLogger(logger_name)
        if level:
            logger.setLevel(getattr(logging, level.upper()))
        logger_handler = logging.handlers.TimedRotatingFileHandler(
            logger_file_path,
            when='d',
            backupCount=3,
            encoding='utf8'
        )
        logger_handler.setFormatter(logging.Formatter('%(asctime)s [%(levelname)s] %(name)s: %(message)s'))
        logger.addHandler(logger_handler)
        self._logger = logger

    def add_event_listener(self, listener):
        self.__listeners.append(listener)

    def is_valid(self):
        for node_id in self.__graph.nodes:
            if isinstance(self.__data[node_id]['object'], PNode):
                try:
                    self.__data[node_id]['object'].check_node_input()
                except AnsibleWorkflowPlaybookNodeCheck as e:
                    self._validation_errors.append(str(e))

        # search cycle
        try:
            cycle = nx.find_cycle(self.__graph)
            error = "The workflow is cyclic: %s" % cycle
            self._logger.error(error)
            self._validation_errors.append(error)
        except nx.NetworkXNoCycle:
            pass

        if self._validation_errors:
            self._logger.error(
                "Impossible to run the workflow due to errors on some playbook"
            )
            return False
        return True

    def get_original_graph(self) -> nx.DiGraph:
        return self.__original_graph

    def get_original_graph_edges(self) -> typing.List[typing.List[str]]:
        return [[u, v] for u, v in self.__original_graph.edges()]

    def add_link(self, node_id: str, next_node_id: str):
        self.__graph.add_edge(node_id, next_node_id)

    def add_node(self, node: Node, other: dict = None):
        node.set_logger(self._logger)

        # check if the node already exists (is auto added if a link is added)
        if (self.is_node_present(node.get_id()) and node.get_id() in self.__data) or ',' in node.get_id():
            if node.get_id() in ['_s', '_e', '_root']:
                msg = "The node id %s name is reserved for internal purpose" % node.get_id()
            elif ',' in node.get_id():
                msg = "The node id %s contains unallowed characters ','" % node.get_id()
            else:
                msg = "Node id %s is already present" % node.get_id()
            self._logger.fatal(msg)
            raise AnsibleWorkflowDuplicateNodeId(msg)

        # store the id of the
        self.__graph.add_node(node.get_id())

        # attach datas to the graph node
        node_data = {}
        if other:
            node_data = other

        # attach an instance of the Node class to the graph node
        node_data.update(dict(object=node))
        self.__data[node.get_id()] = node_data

    def get_node_object(self, node_id: str) -> Node:
        return self.__data[node_id]['object']

    def notify_event(self, event_type: WorkflowEventType,
                     event: typing.Union[NodeStatus, WorkflowStatus],
                     content: typing.Any = None):
        event_obj = WorkflowEvent(event_type, event, content)
        self._logger.debug("Notifying TYPE: %s EVENT: %s CONTENT: %s" %
                           (event_type, event, content))

        for listener in self.__listeners:
            listener.notify_event(event_obj)

    def is_node_present(self, node_id: str):
        if node_id in self.__graph.nodes:
            return True
        return False

    def get_node(self, node_id: str) -> typing.List[typing.Any]:
        return self.__graph.nodes[node_id], self.__data[node_id]

    def get_nodes(self):
        return self.__graph.nodes

    def is_node_runnable(self, node_id):
        self._logger.debug("Check node %s can be run" % node_id)
        in_edges = self.__graph.in_edges(node_id)
        for edge in in_edges:
            previous_node = edge[0]
            self._logger.debug("\tPrevious node %s status: %s" % (previous_node, self.__data[previous_node]['object'].get_status()))
            if self.__data[previous_node]['object'].get_status() not in [NodeStatus.ENDED, NodeStatus.SKIPPED]:
                self._logger.debug("\tNot ended: %s" % previous_node)
                return False
        return True

    def is_running(self):
        return len(self.__running_nodes) != 0

    def get_running_status(self):
        return self.__running_status

    def get_running_nodes(self):
        return self.__running_nodes

    def get_graph(self):
        return self.__graph

    def get_node_datas(self):
        return self.__data

    def run_node(self, node_id):
        node = self.get_node_object(node_id)
        node.run()
        self._logger.info("Node: %s - %s - [ %s - ... ]" % (node, 'starting',
                            node.get_telemetry()["started"]))
        self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.RUNNING, node)

    def skip_node(self, node_id):
        node = self.get_node_object(node_id)
        self._logger.info("Node: %s - %s - [ %s - ... ]" % (node, 'skipped',
                            node.get_telemetry()["started"]))
        self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.SKIPPED, node)

    def add_running_node(self, node_id):
        self.__running_nodes.append(node_id)

    def is_stopping(self):
        return self.__stopped

    def stop(self, mode: str = 'graceful'):
        self._logger.info(f"Stop requested with mode: {mode}")
        self.__stopping = True
        self.__running_status = WorkflowStatus.STOPPING
        self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, f"Workflow stopping ({mode})")
        if mode == 'hard':
            for node_id in self.get_running_nodes():
                node = self.get_node_object(node_id)
                if isinstance(node, PNode):
                    node.stop()
        self.__pause_event.set()

    def pause(self):
        self._logger.info("Pausing workflow")
        self.__running_status = WorkflowStatus.PAUSED
        self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, "Workflow paused")
        self.__pause_event.clear()

    def resume(self):
        self._logger.info("Resuming workflow")
        self.__running_status = WorkflowStatus.RUNNING
        self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, "Workflow resumed")
        self.__pause_event.set()

    def _is_waiting_for_confirmation(self):
        for node_id in self.get_nodes():
            if self.get_node_object(node_id).get_status() == NodeStatus.AWAITING_CONFIRMATION:
                return True
        return False

    def get_some_failed_task(self):
        some_failed_tasks = False
        for node_id in self.get_nodes():
            if self.get_node_object(node_id).get_status() not in [NodeStatus.ENDED, NodeStatus.SKIPPED]:
                # print('--nodeid({}) KO {}'.format(node_id, self.get_node_object(node_id).get_status()))
                some_failed_tasks = True
            else:
                # print('--nodeid({}) ok {}'.format(node_id, self.get_node_object(node_id).get_status()))
                pass
        return some_failed_tasks

    def __run_step(self, end_node="_e"):
        self._logger.debug(f"__run_step: running_nodes={self.__running_nodes}")
        for node_id in list(self.__running_nodes):
            node = self.get_node_object(node_id)
            status = node.get_status()
            self._logger.debug(f"__run_step: processing node {node_id} with status {status}")
            # if current node is ended search for next nodes
            if isinstance(node, CNode) and status == NodeStatus.RUNNING:
                node.set_status(NodeStatus.ENDED)
            elif isinstance(node, (BNode, INode)) and status == NodeStatus.NOT_STARTED:
                node.set_status(NodeStatus.ENDED)
                self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.ENDED, node)
            elif status in [NodeStatus.ENDED, NodeStatus.SKIPPED]:
                self._logger.info(f"Node {node_id} finished with status {status}. Setting end time.")
                self.__running_nodes.remove(node_id)
                if not node.is_skipped():
                    node.set_ended_time(datetime.now())
                    self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.ENDED, node)

                for out_edge in self.__graph.out_edges(node_id):
                    next_node_id = out_edge[1]
                    next_node = self.get_node_object(next_node_id)

                    # check if a node as previous nodes ended and not already started
                    if self.is_node_runnable(next_node_id) and next_node_id not in self.__running_nodes:
                        if next_node_id != end_node and not self.__stopping:
                            self.__running_nodes.append(next_node_id)
                            if isinstance(next_node, PNode):
                                # run a node
                                self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.PRE_RUNNING, next_node)
                                if self.__doubtful_mode:
                                    next_node.set_status(NodeStatus.AWAITING_CONFIRMATION)
                                    self.notify_event(
                                        WorkflowEventType.NODE_EVENT,
                                        NodeStatus.AWAITING_CONFIRMATION,
                                        next_node
                                    )
                                elif not next_node.is_skipped():
                                    self.run_node(next_node_id)
                                else:
                                    self.skip_node(next_node_id)
                            elif isinstance(next_node, CNode):
                                next_node.set_status(NodeStatus.AWAITING_CONFIRMATION)
                                self.notify_event(
                                    WorkflowEventType.NODE_EVENT,
                                    NodeStatus.AWAITING_CONFIRMATION,
                                    next_node
                                )


            elif status == NodeStatus.AWAITING_CONFIRMATION:
                self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.AWAITING_CONFIRMATION, node)
                self.__running_nodes.remove(node_id)
            elif status == NodeStatus.STOPPED:
                self._logger.info(f"Node {node_id} stopped. Setting end time.")
                node.set_ended_time(datetime.now())
                self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.STOPPED, node)
                self.__running_nodes.remove(node_id)
            elif status == NodeStatus.FAILED:
                # just remove a failed node
                # print("Failed node %s" % node_id)
                node.set_ended_time(datetime.now())
                self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.FAILED, node)
                self.__running_nodes.remove(node_id)
                # Do not set workflow status to FAILED here, to allow for retry.

            if isinstance(node, PNode) and node.get_status() in ['ended', 'failed', 'skipped']:
                self._logger.info("Node: %s - %s - [ %s - %s]" % (node_id, node.get_status(),
                                  node.get_telemetry()['started'], node.get_telemetry()['ended']))

    def _set_skipped_nodes(self, start_node: str, end_node: str):
        '''
        Set the nodes to be skipped taking start node and end nodes into account and
        also filtered nodes.
        Args:
            start_node (string): The identifier of the starting node for the graph
            end_node (string): The identifier of the ending node for the graph
        '''
        # set skipped nodes from filtered nodes
        for node in self.__skipped_nodes:
            self.get_node_object(node).set_skipped()
        # skipped from start
        self._logger.info("Setting skipped %s" % self.__graph.in_edges(start_node))
        skipped_from_start = [s for s, _ in self.__graph.in_edges(start_node)]
        while len(skipped_from_start) > 0:
            actual_node_id = skipped_from_start.pop()
            actual_node = self.get_node_object(actual_node_id)
            actual_node.set_skipped()
            for prev, _ in self.__graph.in_edges(actual_node.get_id()):
                skipped_from_start.append(prev)

        skipped_after_end = [e for _, e in self.__graph.out_edges(end_node)]
        while len(skipped_after_end) > 0:
            actual_node_id = skipped_after_end.pop()
            actual_node = self.get_node_object(actual_node_id)
            actual_node.set_skipped()
            for _, next in self.__graph.out_edges(actual_node.get_id()):
                skipped_after_end.append(next)

    def restart_failed_node(self, node_id: str):
        node = self.get_node_object(node_id)
        if not node or node.get_status() != NodeStatus.FAILED:
            self._logger.warning(f"Node {node_id} cannot be restarted.")
            return

        self._logger.info(f"Restarting node {node_id}")

        # Set status back to RUNNING
        self.__running_status = WorkflowStatus.RUNNING
        self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, f"Workflow resuming from node {node_id}")

        # Reset node status
        node.reset_status()

        self.run_node(node_id)
        self.add_running_node(node_id)
        self.__resume_event.set()

    def skip_failed_node(self, node_id: str):
        node = self.get_node_object(node_id)
        if not node or node.get_status() != NodeStatus.FAILED:
            self._logger.warning(f"Node {node_id} cannot be skipped as it is not in FAILED state.")
            return

        self._logger.info(f"Skipping failed node {node_id}")

        # Set status back to RUNNING
        self.__running_status = WorkflowStatus.RUNNING
        self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, f"Workflow resuming, skipping node {node_id}")

        # Set node as skipped
        node.set_skipped()
        self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.SKIPPED, node)

        # Add node to running nodes so its successors can be processed
        self.add_running_node(node_id)
        self.__resume_event.set()

    def approve_node(self, node_id: str):
        node = self.get_node_object(node_id)
        if not node or node.get_status() != NodeStatus.AWAITING_CONFIRMATION:
            self._logger.warning(f"Node {node_id} cannot be approved.")
            return

        self._logger.info(f"Approving node {node_id}")
        node.set_status(None)
        if isinstance(node, CNode):
            node.set_status(NodeStatus.RUNNING)
            self.add_running_node(node_id)
        else:
            self.run_node(node_id)

    def disapprove_node(self, node_id: str):
        node = self.get_node_object(node_id)
        if not node or node.get_status() != NodeStatus.AWAITING_CONFIRMATION:
            self._logger.warning(f"Node {node_id} cannot be disapproved.")
            return

        self._logger.info(f"Disapproving node {node_id}")
        node.set_status(NodeStatus.FAILED)
        self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.FAILED, node)


    def run(self, start_node: str = "_s", end_node: str = "_e", verify_only: bool = False):
        '''
        Run the workflows starting from a graph node until reaching the end node.
        Args:
            start_node (string): The identifier of the starting node for the graph
            end_node (string): The identifier of the ending node for the graph
            verify_only (bool): A flag that skip the workflow run and verify only the correctness

        '''

        # perform validation of the
        if not self.is_valid():
            self.__running_status = WorkflowStatus.FAILED
            self.notify_event(
                WorkflowEventType.WORKFLOW_EVENT,
                self.__running_status,
                self._validation_errors,
            )
            return

        if verify_only:
            self.__running_status = WorkflowStatus.ENDED
            return

        if self.__running_status != WorkflowStatus.NOT_STARTED:
            # This allows to re-enter the loop on retry
            if self.__running_status == WorkflowStatus.RUNNING:
                return
            raise Exception("Already running")

        # check the starting node
        self._logger.info("Start from node %s" % start_node)
        if not self.is_node_present(start_node):
            error = "Starting node not exist: %s" % start_node
            self._logger.error("Starting node not exist: %s" % start_node)
            self.__running_status = WorkflowStatus.FAILED
            self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, error)
            return

        self._set_skipped_nodes(start_node, end_node)
        start_node_object = self.get_node_object(start_node)
        self.__running_status = WorkflowStatus.RUNNING
        self.add_running_node(start_node)
        self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, start_node)

        if isinstance(start_node_object, PNode):
            self.notify_event(WorkflowEventType.NODE_EVENT, NodeStatus.RUNNING, start_node_object)
            self.run_node(start_node)

        # loop over nodes
        while not self.__stopped:
            self.__pause_event.wait()
            self.__run_step(end_node)

            if not self.is_running():
                if self.__stopping:
                    break

                if self._is_waiting_for_confirmation():
                    if self.__running_status != WorkflowStatus.PAUSED:
                        self.__running_status = WorkflowStatus.PAUSED
                        self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, 'Workflow paused, waiting for confirmation.')
                    time.sleep(0.5) # Prevent busy-waiting
                    continue

                if self.get_some_failed_task():
                    # There are failed tasks, set status and wait for user to retry
                    self.__running_status = WorkflowStatus.FAILED
                    self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, 'Workflow failed, waiting for retry.')
                    self.__resume_event.clear()
                    self.__resume_event.wait(timeout=1)
                else:
                    # No running nodes and no failed nodes, we are done
                    self.__running_status = WorkflowStatus.ENDED
                    self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, end_node)
                    break

            time.sleep(0.2)

        if self.__stopping and self.__running_status != WorkflowStatus.ENDED:
            self.__running_status = WorkflowStatus.FAILED
            self.notify_event(WorkflowEventType.WORKFLOW_EVENT, self.__running_status, "Workflow stopped")



import threading
import os
import logging
import logging.handlers
import abc
from .api_client import ApiClient
from ..core.models import WorkflowStatus

class WorkflowOutput(threading.Thread):
    '''
    A general workflow output class to be implemented by subclasses
    '''
    def __init__(self, backend_url, event, logging_dir, log_level, cmd_args):
        threading.Thread.__init__(self)
        self._define_logger(logging_dir, log_level)
        self.api_client = ApiClient(backend_url, logger=self._logger)
        self._refresh_interval = 2
        self.__verify_only = cmd_args.verify_only
        self.__interactive_retry = cmd_args.interactive_retry
        self.event: threading.Event = event

    def is_verify_only(self):
        return self.__verify_only

    def _define_logger(self, logging_dir, level):
        logger_name = self.__class__.__name__
        # Use a fixed log name for now, as we don't have the workflow object here.
        self._log_name = "frontend.log"
        logger_file_path = os.path.join(logging_dir, self._log_name)
        if not os.path.exists(os.path.dirname(logger_file_path)):
            os.makedirs(os.path.dirname(logger_file_path))

        logger = logging.getLogger(logger_name)
        if level:
            logger.setLevel(getattr(logging, level.upper()))
        logger_handler = logging.handlers.TimedRotatingFileHandler(
            logger_file_path,
            when='d',
            backupCount=3,
            encoding='utf8'
        )
        logger_handler.setFormatter(logging.Formatter('%(asctime)s [%(levelname)s] %(name)s: %(message)s'))
        logger.addHandler(logger_handler)
        self._logger = logger
        self._logging_dir = logging_dir

    def run(self):
        self._logger.info("WorkflowOutput run")
        self.draw_init()

        status_data = None
        while not self.event.is_set():
            status_data = self.api_client.get_workflow_status()
            status = status_data.get('status') if status_data else None
            self._logger.info(f"Checking status: {status}")

            if status == WorkflowStatus.ENDED.value:
                break

            if status == WorkflowStatus.FAILED.value and not self.__interactive_retry:
                break

            self.draw_step()

            if hasattr(self, 'user_chose_to_quit') and self.user_chose_to_quit:
                break

            self.draw_pause()

        if not self.event.is_set():
            self._logger.info(f"Final status: {status}. Exiting loop.")
            self.draw_end(status_data=status_data)

    @abc.abstractmethod
    def draw_init(self, *args, **kwargs):
        ''' Draw initialization'''
        pass

    @abc.abstractmethod
    def draw_end(self, status_data: dict = None, *args, **kwargs):
        ''' Draw initialization'''
        pass

    @abc.abstractmethod
    def draw_step(self):
        ''' Draw the workflow'''
        pass

    @abc.abstractmethod
    def draw_pause(self):
        ''' Need to be implemented to pause after a draw step'''
        pass



import os
print(f"Backend CWD: {os.getcwd()}")
import signal
import sys
import threading
import time
from datetime import datetime
from fastapi import FastAPI, HTTPException, BackgroundTasks
from pydantic import BaseModel, Field
from typing import Optional, Dict, List

class StopWorkflowRequest(BaseModel):
    mode: str = "graceful"

from .core.loader import WorkflowYamlLoader
from .core.engine import AnsibleWorkflow
from .core.models import NodeStatus, WorkflowStatus, PNode, INode, CNode
from .core.exceptions import (
    AnsibleWorkflowLoadingError,
    AnsibleWorkflowValidationError,
    AnsibleWorkflowVaultScript,
    AnsibleWorkflowYAMLNotValid,
    AnsibleWorkflowPlaybookNodeCheck,
)
import jinja2

app = FastAPI()

# Global state
workflow_lock = threading.Lock()
current_workflow: Optional[AnsibleWorkflow] = None

class WorkflowStartRequest(BaseModel):
    workflow_file: str
    extra_vars: Dict = Field(default_factory=dict)
    input_templating: Dict = Field(default_factory=dict)
    check_mode: bool = False
    verbosity: int = 0
    start_from_node: Optional[str] = None
    end_to_node: Optional[str] = None
    skip_nodes: List[str] = Field(default_factory=list)
    filter_nodes: List[str] = Field(default_factory=list)
    log_dir: str = "logs"
    log_dir_no_info: bool = False
    log_level: str = "info"
    verify_only: bool = False
    doubtful_mode: bool = False


@app.post("/workflow")
async def start_workflow(request: WorkflowStartRequest, background_tasks: BackgroundTasks):
    global current_workflow
    with workflow_lock:
        if current_workflow and current_workflow.get_running_status() in [WorkflowStatus.RUNNING,
                                                                           WorkflowStatus.PAUSED,
                                                                           WorkflowStatus.ENDED,
                                                                           WorkflowStatus.FAILED]:
            if current_workflow.get_workflow_file() == request.workflow_file:
                return {"status": "reconnected"}
            else:
                raise HTTPException(status_code=409, detail={
                    "message": "A different workflow is already running",
                    "running_workflow_file": current_workflow.get_workflow_file()
                })

        logging_dir = "%s" % request.log_dir
        if not request.log_dir_no_info:
            logging_dir += "/%s_%s" % (os.path.basename(request.workflow_file), datetime.now().strftime("%Y%m%d_%H%M%S"))

        try:
            loader = WorkflowYamlLoader(
                request.workflow_file,
                logging_dir,
                request.log_level,
                request.input_templating,
                request.check_mode,
                request.verbosity,
                request.doubtful_mode,
            )
            aw = loader.parse(request.extra_vars)
            current_workflow = aw
        except (
            AnsibleWorkflowLoadingError,
            jinja2.exceptions.UndefinedError,
            AnsibleWorkflowVaultScript,
            AnsibleWorkflowValidationError,
        ) as e:
            aw = AnsibleWorkflow(
                workflow_file=request.workflow_file,
                logging_dir=logging_dir,
                log_level=request.log_level,
                doubtful_mode=request.doubtful_mode,
            )
            aw.add_validation_error(str(e))
            aw.set_status(WorkflowStatus.FAILED)
            current_workflow = aw
            # Use a 422 status code for validation errors, as this is more specific than a generic 500.
            raise HTTPException(status_code=422, detail={"validation_errors": [str(e)]})

        if request.filter_nodes:
            aw.set_filtered_nodes(request.filter_nodes)
        if request.skip_nodes:
            aw.set_skipped_nodes(request.skip_nodes)

        start_node = request.start_from_node if request.start_from_node else '_s'
        end_node = request.end_to_node if request.end_to_node else '_e'

        background_tasks.add_task(aw.run, start_node=start_node, end_node=end_node, verify_only=request.verify_only)

    return {"status": WorkflowStatus.RUNNING}


@app.get("/workflow")
def get_workflow_status():
    with workflow_lock:
        if not current_workflow:
            return {"status": WorkflowStatus.NOT_STARTED}

        status = current_workflow.get_running_status()
        response = {"status": status}
        if status == WorkflowStatus.FAILED:
            errors = current_workflow.get_validation_errors()
            if errors:
                response["validation_errors"] = errors
        return response


@app.get("/workflow/nodes")
def get_workflow_nodes():
    with workflow_lock:
        if not current_workflow:
            return []

        nodes_data = []
        all_node_datas = current_workflow.get_node_datas()
        for node_id in current_workflow.get_nodes():
            node_obj = current_workflow.get_node_object(node_id)
            status = node_obj.get_status()
            node_info = {
                "id": node_obj.get_id(),
                "status": status.value if hasattr(status, 'value') else status,
                "type": node_obj.get_type(),
            }

            if node_info['type'] == 'block':
                node_data = all_node_datas.get(node_id, {})
                if 'block' in node_data and 'strategy' in node_data['block']:
                    node_info['strategy'] = node_data['block']['strategy']

            if isinstance(node_obj, PNode):
                node_info.update({
                    "playbook": node_obj.get_playbook(),
                    "inventory": node_obj.get_inventory(),
                    "extravars": node_obj.get_extravars(),
                    "description": node_obj.get_description(),
                    "reference": node_obj.get_reference(),
                })
                node_info.update(node_obj.get_telemetry())
            elif isinstance(node_obj, (INode, CNode)):
                node_info.update({
                    "description": node_obj.get_description(),
                    "reference": node_obj.get_reference(),
                })
            nodes_data.append(node_info)
        return nodes_data

@app.get("/workflow/graph")
def get_workflow_graph():
    with workflow_lock:
        if not current_workflow:
            raise HTTPException(status_code=404, detail="Workflow not found.")
        return {"edges": current_workflow.get_original_graph_edges()}

@app.get("/workflow/node/{node_id}/stdout")
def get_node_stdout(node_id: str):
    with workflow_lock:
        if not current_workflow:
            raise HTTPException(status_code=404, detail="Workflow not found.")

        logging_dir = current_workflow.get_logging_dir()
        node_obj = current_workflow.get_node_object(node_id)
        if not isinstance(node_obj, PNode):
            raise HTTPException(status_code=404, detail="Node is not a playbook node.")

        ident = getattr(node_obj, 'ident', node_id)
        stdout_path = os.path.join(logging_dir, ident, "stdout")

        if not os.path.exists(stdout_path):
            return {"stdout": ""}

        with open(stdout_path, "r") as f:
            return {"stdout": f.read()}


@app.post("/workflow/stop")
def stop_workflow(request: StopWorkflowRequest):
    with workflow_lock:
        if not current_workflow or current_workflow.get_running_status() not in [WorkflowStatus.RUNNING, WorkflowStatus.PAUSED]:
            raise HTTPException(status_code=404, detail="No running or paused workflow to stop.")
        current_workflow.stop(request.mode)
    return {"message": "Workflow stopping."}


@app.post("/workflow/pause")
def pause_workflow():
    with workflow_lock:
        if not current_workflow or current_workflow.get_running_status() != WorkflowStatus.RUNNING:
            raise HTTPException(status_code=404, detail="No running workflow to pause.")
        current_workflow.pause()
    return {"message": "Workflow paused."}


@app.post("/workflow/resume")
def resume_workflow():
    with workflow_lock:
        if not current_workflow or current_workflow.get_running_status() != WorkflowStatus.PAUSED:
            raise HTTPException(status_code=404, detail="No paused workflow to resume.")
        current_workflow.resume()
    return {"message": "Workflow resumed."}


@app.post("/workflow/node/{node_id}/restart")
def restart_node(node_id: str):
    with workflow_lock:
        if not current_workflow:
            raise HTTPException(status_code=404, detail="Workflow not found.")

        current_workflow.restart_failed_node(node_id)

    return {"message": f"Node {node_id} restarting."}


@app.post("/workflow/node/{node_id}/skip")
def skip_node(node_id: str):
    with workflow_lock:
        if not current_workflow:
            raise HTTPException(status_code=404, detail="Workflow not found.")

        current_workflow.skip_failed_node(node_id)

    return {"message": f"Node {node_id} skipped."}


@app.post("/workflow/node/{node_id}/approve")
def approve_node(node_id: str):
    with workflow_lock:
        if not current_workflow:
            raise HTTPException(status_code=404, detail="Workflow not found.")
        current_workflow.approve_node(node_id)
    return {"message": f"Node {node_id} approved."}


@app.post("/workflow/node/{node_id}/disapprove")
def disapprove_node(node_id: str):
    with workflow_lock:
        if not current_workflow:
            raise HTTPException(status_code=404, detail="Workflow not found.")
        current_workflow.disapprove_node(node_id)
    return {"message": f"Node {node_id} disapproved."}


@app.post("/shutdown")
def shutdown():
    with workflow_lock:
        if current_workflow:
            if current_workflow.get_running_status() == WorkflowStatus.RUNNING:
                raise HTTPException(status_code=409, detail="Cannot shutdown while a workflow is running.")
            # Tell the workflow thread to stop
            current_workflow.stop()
            # Give the thread a moment to stop
            time.sleep(1.5)

        # This is a simple way to shutdown for this app.
        # In a real production app, a more graceful shutdown mechanism would be needed.
        os.kill(os.getpid(), signal.SIGTERM)

    return {"message": "Shutting down."}


@app.get("/health")
def health_check():
    return {"status": "ok"}


if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8001, log_level="info")



from enum import Enum


class ExitCodes(Enum):
    """ Define the return codes for the application"""
    VALIDATION_ERROR = 1
    VAULT_SCRIPT_ABSENT = 2
    WORKFLOW_NOT_VALID = 3
    WORKFLOW_FILE_TYPE_NOT_SUPPORTED = 4
    YAML_NOT_VALID = 5
    WORKFLOW_FAILED = 6
    PLAYBOOK_WRONG_PARAMETER = 7
    START_NODE_NOT_EXISTS = 8


class AnsibleWorkflowException(Exception):
    """ Generic Workflow Error """


# loading exceptions
class AnsibleWorkflowLoadingError(Exception):
    """ A general class for loading error """


class AnsibleWorkflowYAMLNotValid(AnsibleWorkflowLoadingError):
    """ Workflow YAML is not valid """


class AnsibleWorkflowUnsupportedVersion(AnsibleWorkflowLoadingError):
    """ Unsupported input file format version """


class AnsibleWorkflowValidationError(AnsibleWorkflowLoadingError):
    """ Unsupported content for the input file"""


class AnsibleWorkflowDuplicateNodeId(AnsibleWorkflowLoadingError):
    """ If some workflow nodes are note unique """


class AnsibleWorkflowImportErrors(AnsibleWorkflowLoadingError):
    """ General class for import errors """


class AnsibleWorkflowImportMissingBlock(AnsibleWorkflowImportErrors):
    """ If the imported file do not contains a block """


class AnsibleWorkflowRecursiveImport(AnsibleWorkflowImportErrors):
    """ If the imported file is trying to import itself """


class AnsibleWorkflowConfigurationError(AnsibleWorkflowException):
    """ Misconfiguration of Workflow """


class AnsibleWorkflowVaultScript(AnsibleWorkflowException):
    """ General vault script exception """


class AnsibleWorkflowVaultScriptNotSet(AnsibleWorkflowVaultScript):
    """ Vault script not present """


class AnsibleWorkflowVaultScriptNotExists(AnsibleWorkflowVaultScript):
    """ Vault script not present """


class AnsibleWorkflowPlaybookNodeCheck(AnsibleWorkflowConfigurationError):
    """ Playbook node check error """



import time
import threading
from datetime import datetime
from rich.console import Console
import sys
import select
import tty
import termios
from rich.table import Table
from rich.prompt import Prompt
from rich.text import Text
from .base import WorkflowOutput
from ..core.models import NodeStatus


class StdoutWorkflowOutput(WorkflowOutput):
    _log_name = 'console.log'

    def __init__(self, backend_url, event, logging_dir, log_level, cmd_args):
        super().__init__(backend_url, event, logging_dir, log_level, cmd_args)
        self._refresh_interval = 2
        self.__console = Console()
        self.__interactive_retry = cmd_args.interactive_retry
        self.__doubtful_mode = cmd_args.doubtful_mode
        self.known_nodes = {}
        self.user_chose_to_quit = False
        self.declined_retry_nodes = set()
        self.approved_nodes = set()
        self.console_lock = threading.Lock()
        self.stop_requested = False

    def draw_init(self):
        self._logger.debug("Initializing stdout output")
        if self.is_verify_only():
            self.__console.print("[bold yellow]Running in VERIFY ONLY mode[/]", justify="center")
        self.__console.print("\n[bold cyan]Press Ctrl+X to stop the workflow or Ctrl+C to detach from the backend[/]\n", justify="center")
        self.__console.print("[italic]Waiting for workflow to start...[/]", justify="center")

        nodes = self.api_client.get_all_nodes()
        while not nodes:
            time.sleep(1)
            nodes = self.api_client.get_all_nodes()

        # calculate first column size
        maximum_first_colum_width = 0
        if nodes:
            for node in nodes:
                if len(node['id']) > maximum_first_colum_width:
                    maximum_first_colum_width = len(node['id'])

        if maximum_first_colum_width < 17:
            self.__first_column_width = 17
        else:
            self.__first_column_width = maximum_first_colum_width

        table = Table(title="Workflow nodes")
        table.add_column("Node", justify="left", style="cyan", no_wrap=True, width=self.__first_column_width)
        table.add_column("Playbook", style="bright_magenta")
        table.add_column("Ref.", style="cyan")
        table.add_column("Started", style="green")
        table.add_column("Ended", style="green")
        table.add_column("Status")


        if nodes:
            for node in nodes:
                node_type = node.get('type')
                if node_type in ['playbook', 'info', 'checkpoint']:
                    self.known_nodes[node['id']] = node

                playbook_col = "-"
                if node_type == 'playbook':
                    playbook_col = node.get('playbook', '-')
                elif node_type == 'info':
                    playbook_col = f"[dim]({node.get('description', 'Info')})[/dim]"
                elif node_type == 'checkpoint':
                    playbook_col = f"[dim]({node.get('description', 'Checkpoint')})[/dim]"

                if node_type in ['playbook', 'info', 'checkpoint']:
                    table.add_row(
                        node['id'],
                        playbook_col,
                        node.get('reference', '-'),
                        node.get('started', ''),
                        node.get('ended', ''),
                        self._render_status(node['status'])
                    )
        self.__console.print(table)
        self.__console.print("")


        if nodes and self.__interactive_retry:
            for node in nodes:
                if node['status'] == NodeStatus.FAILED.value:
                    if node['id'] not in self.declined_retry_nodes:
                        self.handle_retry(node)

        self.__console.print("[italic]Running[/] ...", justify="center")

    def draw_step(self):
        nodes = self.api_client.get_all_nodes()
        found_failed_node_to_prompt = False
        if nodes:
            for node in nodes:
                node_id = node['id']
                if node_id in self.known_nodes and self.known_nodes[node_id]['status'] != node['status']:
                    self.print_node_status_change(node)
                    self.known_nodes[node_id] = node

                if node['status'] == NodeStatus.FAILED.value and self.__interactive_retry and node.get('type') != 'checkpoint':
                    if node['id'] not in self.declined_retry_nodes:
                        self.handle_retry(node)
                        found_failed_node_to_prompt = True

                if node['status'] == NodeStatus.AWAITING_CONFIRMATION.value:
                    if node['id'] not in self.approved_nodes:
                        if node.get('type') == 'checkpoint':
                            if self.handle_checkpoint_node(node):
                                return
                        elif self.__doubtful_mode:
                            if self.handle_doubtful_node(node):
                                return

        status_data = self.api_client.get_workflow_status()
        if status_data.get('status') == 'failed' and not found_failed_node_to_prompt:
            self.user_chose_to_quit = True


    def draw_pause(self):
        ''' Non blocking thread wait'''
        time.sleep(self._refresh_interval)

    def draw_end(self, status_data: dict = None):
        if status_data:
            errors = status_data.get('validation_errors')
            if errors:
                self.__console.print("\n[bold red]Workflow validation failed with errors:[/bold red]")
                for error in errors:
                    self.__console.print(f"- {error}")
                self.__console.print("")

        nodes = self.api_client.get_all_nodes()
        table = Table(title="Running recap")

        table.add_column("Node", justify="left", style="cyan", no_wrap=True, width=self.__first_column_width)
        table.add_column("Playbook", style="bright_magenta")
        table.add_column("Ref.", style="cyan")
        table.add_column("Started", style="green")
        table.add_column("Ended", style="green")
        table.add_column("Status")

        if nodes:
            for node in nodes:
                node_type = node.get('type')
                playbook_col = "-"
                if node_type == 'playbook':
                    playbook_col = node.get('playbook', '-')
                elif node_type == 'info':
                    playbook_col = f"[dim]({node.get('description', 'Info')})[/dim]"
                elif node_type == 'checkpoint':
                    playbook_col = f"[dim]({node.get('description', 'Checkpoint')})[/dim]"

                if node_type in ['playbook', 'info', 'checkpoint']:
                    table.add_row(
                        node['id'],
                        playbook_col,
                        node.get('reference', '-'),
                        node.get('started', ''),
                        node.get('ended', ''),
                        self._render_status(node['status'])
                    )
        self.__console.print(table)
        self.__console.print("")
        self._logger.debug("stdout output ends")

    def _render_status(self, status):
        if status == NodeStatus.RUNNING.value:
            return '[yellow]started[/]'
        elif status == NodeStatus.ENDED.value:
            return '[green]completed[/]'
        elif status == NodeStatus.FAILED.value:
            return '[bright_red]failed[/]'
        elif status == NodeStatus.NOT_STARTED.value:
            return '[white]not started[/]'
        elif status == NodeStatus.SKIPPED.value:
            return '[cyan]skipped[/]'
        elif status == NodeStatus.STOPPED.value:
            return '[red]stopped[/]'
        elif status == NodeStatus.AWAITING_CONFIRMATION.value:
            return '[bold yellow]awaiting confirmation[/]'
        else:
            return 'unknown'

    def handle_doubtful_node(self, node):
        y_or_n = ''
        self.__console.line()
        self.__console.rule("node \[[italic]" + node['id'] +"[/italic]] awaiting confirmation")
        table = Table(show_header=False, show_footer=False, show_lines=False, show_edge=False)
        table.add_column(width=(self.__first_column_width+1), justify="right")
        table.add_column()
        table.add_row('[bright_magenta]Node[/]',f"[cyan]{node['id']}[/]")
        table.add_row('[bright_magenta]Reference[/]',node.get('reference', '-'))
        table.add_row('[bright_magenta]Description[/]',node.get('description', '-'))
        self.__console.print(table)

        while y_or_n.lower() not in ['y', 'n']:
            table = Table(show_header=False, show_footer=False, show_lines=False, show_edge=False)
            table.add_column(width=self.__first_column_width)
            table.add_column(justify="right")
            table.add_column()
            self.__console.print(table)
            self.__console.line()
            y_or_n = Prompt.ask("[white] Do you want to run the node \[{}]? [green]y[/](yes) / [bright_red]n[/](no/skip)".format(node['id']),
                                console=self.__console,
                                show_choices=False,
                                choices=["n","y"])

        self.__console.line()
        self.__console.rule()

        if y_or_n.strip().lower() == 'y':
            self.api_client.approve_node(node['id'])
        elif y_or_n.strip().lower() == 'n':
            self.api_client.disapprove_node(node['id'])

        self.approved_nodes.add(node['id'])
        return True

    def handle_checkpoint_node(self, node):
        y_or_n = ''
        self.__console.line()
        self.__console.rule(f"Checkpoint Reached: [italic]{node['id']}[/italic]")

        description = node.get('description', 'Do you want to proceed?')
        if node.get('reference'):
            description += f"\n[dim]Reference: {node.get('reference')}[/dim]"

        self.__console.print(description, justify="center")

        while y_or_n.lower() not in ['y', 'n']:
            y_or_n = Prompt.ask("[white]Do you want to continue? [green]y[/](yes) / [bright_red]n[/](no/skip)",
                                console=self.__console,
                                show_choices=False,
                                choices=["n","y"])

        self.__console.line()
        self.__console.rule()

        if y_or_n.strip().lower() == 'y':
            self.api_client.approve_node(node['id'])
        elif y_or_n.strip().lower() == 'n':
            self.api_client.disapprove_node(node['id'])

        self.approved_nodes.add(node['id'])

    def print_node_status_change(self, node):
        node_type = node.get('type')
        status = node.get('status')
        timestamp = node.get('ended', '')

        if not timestamp:
             timestamp = datetime.now().strftime('%H:%M:%S')

        table = Table(show_header=False, show_footer=False, show_lines=False, show_edge=False)
        table.add_column(width=(self.__first_column_width +1), justify="right")
        table.add_column()

        message = ""
        if node_type == 'info' and status == NodeStatus.ENDED.value:
            message = f"[bold cyan]INFO:[/] [cyan]{node.get('description', node['id'])}[/]"
        else:
            status_text = self._render_status(node['status'])
            message = f"Node [cyan]{node['id']}[/] is {status_text}"

        table.add_row(timestamp, message)
        self.__console.print(table)

    def handle_retry(self, node):
        y_or_n = ''
        self.__console.line()
        self.__console.rule("node \[[italic]" + node['id'] +"[/italic]] failed")
        table = Table(show_header=False, show_footer=False, show_lines=False, show_edge=False)
        #table.add_column()
        table.add_column(width=(self.__first_column_width+1), justify="right")
        table.add_column()
        table.add_row('[bright_magenta]Node[/]',f"[cyan]{node['id']}[/]")
        table.add_row('[bright_magenta]Reference[/]',node.get('reference', '-'))
        table.add_row('[bright_magenta]Description[/]',node.get('description', '-'))
        self.__console.print(table)

        while y_or_n.lower() not in ['y', 'n', 's', 'l']:
            table = Table(show_header=False, show_footer=False, show_lines=False, show_edge=False)
            table.add_column(width=self.__first_column_width)
            table.add_column(justify="right")
            table.add_column()
            self.__console.print(table)
            self.__console.line()
            y_or_n = Prompt.ask("[white] Do you want to restart the node \[{}]? [green]y[/](yes) / [bright_red]n[/](no) / [cyan]s[/](skip) / [bright_magenta]l[/](logs)".format(node['id']),
                                console=self.__console,
                                show_choices=False,
                                choices=["n","y","s","l"])

            if y_or_n == 'l':
                stdout = self.api_client.get_node_stdout(node['id'])
                if stdout:
                    self.__console.line()
                    self.__console.print(Text.from_ansi(stdout))
        self.__console.line()
        self.__console.rule()

        if y_or_n == 'y':
            self.api_client.restart_node(node['id'])
        elif y_or_n == 's':
            self.api_client.skip_node(node['id'])
        elif y_or_n == 'n':
            self.declined_retry_nodes.add(node['id'])

    def _request_stop(self):
        self.stop_requested = True

    def _handle_stop_request(self):
        self.api_client.pause_workflow()
        with self.console_lock:
            self.__console.print("\n")
            self.__console.print("[bold yellow]Stop workflow requested.[/]")
            self.__console.print("[bold yellow]Choose stop mode[/]: \\[g][dark_orange]raceful[/], \\[h][red]ard[/], or \\[c][cyan]ancel[/]?")
            choice = self.__console.input("> ")
            self._logger.info(f"User chose: {choice}")
            if choice.lower() == 'g':
                self.api_client.stop_workflow(mode="graceful")
                self.__console.print("[yellow]Graceful stop requested.[/]")
            elif choice.lower() == 'h':
                self.api_client.stop_workflow(mode="hard")
                self.__console.print("[red]Hard stop requested.[/]")
            else:
                self.api_client.resume_workflow()
                self.__console.print("[green]Stop request canceled.[/]")
        self.stop_requested = False

    def run(self):
        self._logger.info("WorkflowOutput run")
        self.draw_init()

        is_tty = sys.stdin.isatty()
        if is_tty:
            old_settings = termios.tcgetattr(sys.stdin)

        try:
            if is_tty:
                tty.setcbreak(sys.stdin.fileno())

            status_data = None
            while not self.event.is_set():
                if is_tty and select.select([sys.stdin], [], [], 0) == ([sys.stdin], [], []):
                    c = sys.stdin.read(1)
                    if c == '\x18': # Ctrl+X
                        self._request_stop()

                if self.stop_requested:
                    self._handle_stop_request()

                status_data = self.api_client.get_workflow_status()
                status = status_data.get('status') if status_data else None
                self._logger.info(f"Checking status: {status}")

                if status == "ended":
                    break

                if status == "failed" and not self._WorkflowOutput__interactive_retry:
                    break

                self.draw_step()

                if hasattr(self, 'user_chose_to_quit') and self.user_chose_to_quit:
                    break

                self.draw_pause()
        finally:
            if is_tty:
                termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_settings)


        if not self.event.is_set():
            self._logger.info(f"Final status: {status}. Exiting loop.")
            self.draw_end(status_data=status_data)



from enum import Enum
import typing
from datetime import datetime
import abc
import os
import logging
import ansible_runner
from .exceptions import AnsibleWorkflowPlaybookNodeCheck


class WorkflowStatus(Enum):
    """ Define the character for the application"""
    NOT_STARTED = 'not_started'
    RUNNING = 'running'
    PAUSED = 'paused'
    STOPPING = 'stopping'
    ENDED = 'ended'
    FAILED = 'failed'


class NodeStatus(Enum):
    """ Define the character for the application"""
    RUNNING = 'running'
    PRE_RUNNING = 'pre_running'
    AWAITING_CONFIRMATION = 'awaiting_confirmation'
    ENDED = 'ended'
    FAILED = 'failed'
    SKIPPED = 'skipped'
    NOT_STARTED = 'not_started'
    STOPPED = 'stopped'


class Node():
    ''' An abstract Node class of the graph'''
    __metaclass__ = abc.ABCMeta

    def __init__(self, id: str):
        '''
        Initialize a generic node of the graph
        Args:
            id (str): The unique identifier of the node inside the workflows
        Raises:
            AnsibleWorkflowVaultScriptNotSet: If a node specify some vault ids but the vault script is not set
        '''
        self.__id = id
        self._logger: logging.Logger = logging
        self._started_time: datetime = None
        self._ended_time: datetime = None
        self.__skipped = False
        self._status: typing.Optional[NodeStatus] = None

    def get_id(self) -> str:
        return self.__id

    def set_logger(self, logger: logging.Logger):
        self._logger = logger

    def set_skipped(self):
        self.__skipped = True

    def is_skipped(self):
        return self.__skipped

    def set_status(self, status: NodeStatus):
        self._status = status

    def __eq__(self, other):
        return self.__id == other.get_id()

    def __hash__(self):
        return hash(self.__id)

    def __str__(self):
        return "%s[%s]" % (self.get_type(), self.get_id())

    @abc.abstractmethod
    def get_status(self):
        return NodeStatus.ENDED

    @abc.abstractmethod
    def get_type(self):
        pass

    def set_ended_time(self, time):
        self._ended_time = time

    def set_started_time(self, time):
        self._started_time = time

    def get_telemetry(self):
        return dict(started=self._started_time.strftime("%H:%M:%S") if self._started_time else '',
                    ended=self._ended_time.strftime("%H:%M:%S") if self._ended_time else '')


class BNode(Node):
    def get_status(self):
        if self._status:
            return self._status
        return NodeStatus.NOT_STARTED

    def get_type(self):
        return 'block'


class PNode(Node):
    def __init__(self, id, playbook, inventory, artifact_dir, limit=None, project_path=None, extra_vars={}, vault_ids=[], check_mode=False, diff_mode=True, verbosity=1, description='', reference=''):
        super(PNode, self).__init__(id)
        self.__playbook = playbook
        self.__inventory = inventory
        self.__extravars = extra_vars
        self.__artifact_dir = artifact_dir
        self.__limit = limit
        self.__vault_ids = vault_ids
        self.__project_path = project_path
        self.__thread = None
        self.__runner = None
        self.__check_mode = check_mode
        self.__diff_mode = diff_mode
        self.__verbosity = verbosity
        self.__description = description
        self.__reference = reference
        self.__hard_stop = False

    def check_node_input(self):
        # convert project path in absolute path
        if not os.path.isabs(self.__project_path):
            self.__project_path = os.path.abspath(self.__project_path)

        if not os.path.exists(self.__project_path):
            raise AnsibleWorkflowPlaybookNodeCheck(
                "Node %s project path %s doesn't exists" % (self.get_id(), self.__project_path)
            )

        if self.__inventory is None:
            raise AnsibleWorkflowPlaybookNodeCheck("Node %s inventory not set" % self.get_id())
        elif not os.path.exists(self.__inventory):
            raise AnsibleWorkflowPlaybookNodeCheck(
                "Node %s inventory doesn't exists: %s" % (self.get_id(), self.__inventory)
            )
        else:
            self.__inventory = os.path.abspath(self.__inventory)

        if self.__project_path and not os.path.isabs(self.__playbook):
            self.__playbook = os.path.join(self.__project_path, self.__playbook)

        if not os.path.exists(self.__playbook):
            raise AnsibleWorkflowPlaybookNodeCheck(
                "Node %s playbook doesn't exists: %s" % (self.get_id(), self.__playbook)
            )

    def get_verbosity(self):
        return self.__verbosity

    def set_verbosity(self, verbosity):
        self.__verbosity = verbosity

    def get_status(self):
        if self._status:
            return self._status
        if self.is_skipped():
            return NodeStatus.SKIPPED
        elif self.__thread is None:
            return NodeStatus.NOT_STARTED
        else:
            # print("Node %s status is %s - error is %s" % (self.get_id(), self.__runner.errored, self.__runner.status ))
            if self.__thread.is_alive():
                return NodeStatus.RUNNING
            elif self.is_canceled():
                return NodeStatus.STOPPED
            elif self.is_failed():
                return NodeStatus.FAILED
            else:
                return NodeStatus.ENDED

    def get_type(self):
        return 'playbook'

    def is_canceled(self):
        return self.__runner.status == 'canceled'

    def is_failed(self):
        return self.__runner.status == 'failed'

    def get_playbook(self):
        return self.__playbook

    def get_inventory(self):
        return self.__inventory

    def get_extravars(self):
        return self.__extravars

    def get_description(self):
        return self.__description

    def get_reference(self):
        return self.__reference

    def stop(self):
        self._logger.info("Stopping node %s" % self.get_id())
        self.__hard_stop = True

    def _cancel_callback(self):
        return self.__hard_stop

    def reset_status(self):
        self.__thread = None
        self.__runner = None
        self._started_time = None
        self._ended_time = None

    def run(self):
        self.set_started_time(datetime.now())
        self.__inventory = os.path.abspath(self.__inventory)
        self.__playbook = os.path.abspath(self.__playbook)

        # put the current directory to the parent of the playbook
        env_vars = {}

        if self.__project_path:
            env_vars = {'ANSIBLE_COLLECTIONS_PATHS': os.path.join(self.__project_path, 'collections')}

        playbook_cmd_line = ' '.join(["--vault-id %s" % vid for vid in self.__vault_ids])
        if self.__check_mode:
            playbook_cmd_line += ' --check'
        if self.__diff_mode:
            playbook_cmd_line += ' --diff'

        # modify identification in case of multiple start
        ident = self.get_id()
        if os.path.exists(os.path.join(self.__artifact_dir, "%s" % self.get_id())):
            i=1
            ident = "%s_%s" % (self.get_id(), i)
            while os.path.exists(os.path.join(self.__artifact_dir, "%s_%s" % (self.get_id(), i))):
                i = i + 1
                ident = "%s_%s" % (self.get_id(), i)
        self.ident = ident
        self.__thread, self.__runner = ansible_runner.run_async(playbook=self.__playbook,
                                                                inventory=self.__inventory,
                                                                ident=ident,
                                                                limit=self.__limit,
                                                                project_dir=self.__project_path,
                                                                event_handler=lambda x: False,
                                                                omit_event_data=True,
                                                                envvars=env_vars,
                                                                verbosity=self.get_verbosity(),
                                                                artifact_dir=self.__artifact_dir,
                                                                settings={
                                                                    'suppress_ansible_output': True
                                                                },
                                                                cancel_callback=self._cancel_callback,
                                                                # vault_ids=self.__vault_ids,
                                                                cmdline=playbook_cmd_line,
                                                                extravars=self.__extravars,
                                                                quiet=True)


class INode(Node):
    def __init__(self, id, description='', reference=''):
        super(INode, self).__init__(id)
        self.__description = description
        self.__reference = reference

    def get_status(self):
        if self._status:
            return self._status
        return NodeStatus.NOT_STARTED

    def get_type(self):
        return 'info'

    def get_description(self):
        return self.__description

    def get_reference(self):
        return self.__reference


class CNode(Node):
    def __init__(self, id, description='', reference=''):
        super(CNode, self).__init__(id)
        self.__description = description
        self.__reference = reference

    def get_status(self):
        if self._status:
            return self._status
        return NodeStatus.NOT_STARTED

    def get_type(self):
        return 'checkpoint'

    def get_description(self):
        return self.__description

    def get_reference(self):
        return self.__reference


class WorkflowEventType(Enum):
    """ Define the return codes for the application"""
    NODE_EVENT = 1
    WORKFLOW_EVENT = 2


class WorkflowEvent(object):
    def __init__(self, event_type: WorkflowEventType,
                 event: typing.Union[NodeStatus, WorkflowStatus],
                 content: typing.Any = None):
        self._type = event_type
        self._event = event
        self._content = content

    def get_type(self) -> WorkflowEventType:
        return self._type

    def get_event(self):
        return self._event, self._content

    def __str__(self):
        return '%s -> %s: %s' % (self._type.name, self._event, self._content)


class WorkflowListener(object):
    @abc.abstractmethod
    def notify_event(self, event: WorkflowEvent):
        ''' A notification method to be overwrited'''
        pass



import jsonschema
import json
from .exceptions import AnsibleWorkflowValidationError

def validate_workflow(instance, schema_path):
    with open(schema_path, 'r') as f:
        schema = json.load(f)
    jsonschema.validate(instance=instance, schema=schema)



import os
import time
import threading
import itertools
from itertools import cycle
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", message="networkx backend defined more than once: nx-loopback")
    import networkx as nx
from rich.highlighter import Highlighter
from rich.text import Text
from rich.pretty import Pretty
from textual.app import App, ComposeResult
from textual.widgets import Header, Footer, Static, Tree, RichLog, DataTable, Button, Label
from textual.containers import Horizontal, Vertical, Container
from textual.screen import Screen, ModalScreen
from textual import work
from textual.reactive import reactive
from textual.theme import BUILTIN_THEMES
from collections import deque
from textual.css.query import NoMatches
from .base import WorkflowOutput
from ..core.models import NodeStatus
from .api_client import ApiClient


class QuitScreen(ModalScreen):
    """Screen with a dialog to quit."""

    def compose(self) -> ComposeResult:
        yield Container(
            Label("Are you sure you want to quit?", id="question"),
            Horizontal(
                Button("Yes", variant="error", id="quit"),
                Button("No", variant="primary", id="cancel"),
                id="buttons",
            ),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "quit":
            self.dismiss(True)
        else:
            self.dismiss(False)


class StopWorkflowScreen(ModalScreen):
    """Screen with a dialog to stop the workflow."""

    def __init__(self, running_nodes: list = None, **kwargs):
        super().__init__(**kwargs)
        self.running_nodes = running_nodes or []

    def compose(self) -> ComposeResult:
        dialog_children = [
            Label("Are you sure you want to stop the workflow?", id="question")
        ]

        if self.running_nodes:
            table = DataTable(id="running_nodes_table")
            table.add_column("Running Nodes")
            for node in self.running_nodes:
                table.add_row(node)
            dialog_children.extend([
                Label("\nThe following nodes are currently running:"),
                table
            ])

        dialog_children.append(
            Horizontal(
                Button("Graceful Stop", variant="primary", id="graceful_stop"),
                Button("Hard Stop", variant="error", id="hard_stop"),
                Button("No", variant="primary", id="cancel"),
                id="buttons",
            )
        )

        yield Vertical(*dialog_children, id="dialog")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "graceful_stop":
            self.dismiss("graceful")
        elif event.button.id == "hard_stop":
            self.dismiss("hard")
        else:
            self.dismiss(None)


class DoubtfulNodeScreen(ModalScreen):
    """Screen with a dialog to approve or skip a node."""

    def __init__(self, node_id: str, message: str, **kwargs):
        super().__init__(**kwargs)
        self.node_id = node_id
        self.message = message

    def compose(self) -> ComposeResult:
        yield Container(
            Label(self.message, id="question"),
            Horizontal(
                Button("Approve", variant="success", id="approve"),
                Button("Skip", variant="primary", id="skip"),
                id="buttons",
            ),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "approve":
            self.dismiss(True)
        else:
            self.dismiss(False)


class NullHighlighter(Highlighter):
    def highlight(self, text):
        pass

class TextualWorkflowOutput(WorkflowOutput):
    _log_name = 'textual.log'

    def __init__(self, backend_url, event, logging_dir, log_level, cmd_args):
        # We don't call super().__init__ because Textual has its own way of running.
        self._define_logger(logging_dir, log_level)
        self.api_client = ApiClient(backend_url, logger=self._logger)
        self.cmd_args = cmd_args
        self._WorkflowOutput__verify_only = cmd_args.verify_only
        self.app = self.WorkflowApp(self, cmd_args)

    def run(self):
        """
        This method is called directly from __main__.py for textual mode.
        It launches the Textual app.
        """
        self.app.run()

    # The following methods are not used in Textual mode as the app handles the loop.
    def draw_init(self): pass
    def draw_end(self, status_data: dict = None): pass
    def draw_step(self): pass
    def draw_pause(self): pass


    class WorkflowApp(App):
        CSS_PATH = "style.css"
        BINDINGS = [
            ("t", "cycle_themes", "Cycle Themes"),
            ("x", "request_stop_workflow", "Stop Workflow"),
            ("q", "request_quit", "Quit")
        ]

        status_message = reactive("Connecting to backend...")

        def __init__(self, outer_instance, cmd_args):
            super().__init__()
            self.outer_instance = outer_instance
            self.workflow_filename = os.path.basename(cmd_args.workflow)
            if self.outer_instance.is_verify_only():
                self.title = f"Workflow Viewer (Verify Only)"
            else:
                self.title = "Workflow Viewer"

            self.api_client = outer_instance.api_client
            self.selected_node_id = None
            self.theme_cycle = cycle(BUILTIN_THEMES.keys())
            self.theme = "gruvbox"
            # Advance the cycle to the default theme
            while next(self.theme_cycle) != self.theme:
                pass
            self.tree_nodes = {}
            self.node_data = {}
            self.graph = nx.DiGraph()
            self.spinner_icons = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
            self.approved_nodes = set()
            self.status_icons = {
                NodeStatus.NOT_STARTED.value: "○",
                NodeStatus.PRE_RUNNING.value: "[yellow]…[/yellow]",
                NodeStatus.RUNNING.value: "[yellow]○[/yellow]",
                NodeStatus.AWAITING_CONFIRMATION.value: "[bold yellow]?[/]",
                NodeStatus.ENDED.value: "[green]✔[/green]",
                NodeStatus.FAILED.value: "[red]✖[/red]",
                NodeStatus.SKIPPED.value: "[cyan]»[/cyan]",
                NodeStatus.STOPPED.value: "[red]■[/red]",
            }
            # This dictionary now only serves as a flag to indicate if a spinner worker
            # has been started for a node, to prevent duplicates.
            self.active_spinners = set()
            self.stdout_watcher = None
            self._shutdown_event = threading.Event()
            self.action_buttons = None
            self.stdout_log = None
            self.details_table = None
            self._node_tree = None
            self.doubtful_node_queue = deque()
            self.pending_confirmation_nodes = set()

        def compose(self) -> ComposeResult:
            yield Header()
            with Horizontal():
                yield Tree(self.workflow_filename, id="workflow_tree", classes="sidebar")
                with Vertical():
                    yield DataTable(id="node_details", show_cursor=False, show_header=False)
                    with Horizontal(id="action_buttons"):
                        yield Button("Relaunch", id="relaunch_button", variant="success")
                        yield Button("Skip", id="skip_button", variant="error")
                    playbook_stdout_log = RichLog(id="playbook_stdout", markup=False, highlight=True)
                    playbook_stdout_log.highlighter = NullHighlighter()
                    yield playbook_stdout_log
            yield Static("Connecting to backend...", id="status_bar")
            yield Footer()

        def watch_status_message(self, message: str) -> None:
            try:
                status_bar = self.query_one("#status_bar", Static)
                status_bar.update(message)
            except NoMatches:
                pass

        @work(thread=True)
        def update_status(self):
            if self.api_client.check_health():
                self.status_message = "[green]Backend: Connected[/green]"

                workflow_status = self.api_client.get_workflow_status()
                if workflow_status and workflow_status.get('status') == 'failed':
                    errors = workflow_status.get('validation_errors')
                    if errors:
                        self.status_message = f"[bold red]Validation Error:[/bold red] {errors[0]}"
            else:
                self.status_message = "[red]Backend: Disconnected[/red]"
                self.call_from_thread(self._set_widget_display, self.action_buttons, False)

        def on_mount(self) -> None:
            self.action_buttons = self.query_one("#action_buttons")
            self.stdout_log = self.query_one("#playbook_stdout", RichLog)
            self.details_table = self.query_one("#node_details", DataTable)
            self._node_tree = self.query_one(Tree)
            self.initial_setup()
            self.set_interval(1, self.update_status)
            self.set_interval(0.5, self.update_node_statuses)

        def action_quit(self) -> None:
            """Called when the user quits the application."""
            self._shutdown_event.set()
            self.exit()

        def action_request_quit(self) -> None:
            """Action to display the quit dialog."""
            self.push_screen(QuitScreen(), self.check_quit)

        def action_request_stop_workflow(self) -> None:
            """Action to display the stop workflow dialog."""
            self.api_client.pause_workflow()
            running_nodes = self.get_running_nodes()
            self.push_screen(StopWorkflowScreen(running_nodes=running_nodes), self.check_stop_workflow)

        def action_stop_workflow(self, mode: str) -> None:
            """Action to stop the workflow."""
            self.api_client.stop_workflow(mode)

        def check_quit(self, should_quit: bool) -> None:
            """Called when the QuitScreen is dismissed."""
            if should_quit:
                self.action_quit()

        def check_stop_workflow(self, stop_mode: str) -> None:
            """Called when the StopWorkflowScreen is dismissed."""
            self.outer_instance._logger.info(f"Stop confirmation dismissed with mode: {stop_mode}")
            if stop_mode:
                self.action_stop_workflow(stop_mode)
            else:
                self.api_client.resume_workflow()

        def check_doubtful_node(self, result: bool, node_id: str) -> None:
            """Called when the DoubtfulNodeScreen is dismissed."""
            if result:
                self.api_client.approve_node(node_id)
            else:
                self.api_client.disapprove_node(node_id)
            self.approved_nodes.add(node_id)
            self.pending_confirmation_nodes.remove(node_id)
            self._process_doubtful_queue()

        def _set_widget_display(self, widget, display):
            widget.display = display

        def _push_doubtful_node_screen(self, node_id: str, message: str):
            self.push_screen(
                DoubtfulNodeScreen(node_id, message),
                lambda result: self.check_doubtful_node(result, node_id)
            )

        def _process_doubtful_queue(self):
            if not self.is_modal and self.doubtful_node_queue:
                node_id, message = self.doubtful_node_queue.popleft()
                self._push_doubtful_node_screen(node_id, message)

        def action_cycle_themes(self) -> None:
            """An action to cycle themes."""
            self.app.theme = next(self.theme_cycle)

        def get_running_nodes(self):
            running_nodes = []
            for node_id, node_data in self.node_data.items():
                if node_data.get('status') == NodeStatus.RUNNING.value:
                    running_nodes.append(node_id)
            return running_nodes

        @work(thread=True)
        def initial_setup(self):
            # Fetch graph and node data once
            edges = self.api_client.get_workflow_graph()
            if edges is not None:
                self.graph.add_edges_from(edges)
            if "_root" not in self.graph:
                self.graph.add_node("_root")

            nodes = self.api_client.get_all_nodes()
            if nodes is not None:
                for node in nodes:
                    self.node_data[node['id']] = node

            # Build the tree
            root_node_id = "_root"

            def build_initial_tree():
                root_node = self._node_tree.root
                root_node.data = root_node_id
                self.tree_nodes[root_node_id] = root_node
                self._build_tree(root_node_id, root_node)
                self._node_tree.root.expand_all()

            self.call_from_thread(build_initial_tree)

        def _build_tree(self, node_id, tree_node):
            for child_id in self.graph.successors(node_id):
                if child_id in ['_s', '_e']:
                    continue

                child_node_data = self.node_data.get(child_id, {})
                node_type = child_node_data.get('type')

                allow_expand = node_type == 'block'
                if node_type == 'block':
                    label = f"[b]{child_id}[/b]"
                elif node_type == 'info':
                    label = f"[cyan]i[/] {child_id}"
                else:
                    icon = self.status_icons.get(child_node_data.get('status'), " ")
                    label = f"{icon} {child_id}"

                child_tree_node = tree_node.add(label, data=child_id, allow_expand=allow_expand)
                self.tree_nodes[child_id] = child_tree_node

                if self.graph.out_degree(child_id) > 0:
                    self._build_tree(child_id, child_tree_node)

        @work(thread=True, exclusive=True)
        def update_node_statuses(self):
            # Sanitize the data from the API to prevent processing duplicate statuses
            nodes_from_api = self.api_client.get_all_nodes()
            if nodes_from_api is None:
                return
            final_node_states = {node['id']: node for node in nodes_from_api}

            nodes_need_approval = False
            for node_id, node in final_node_states.items():
                if node_id in self.tree_nodes and node_id != "_root":
                    # Update the central data store
                    self.node_data[node_id] = node

                    tree_node = self.tree_nodes[node_id]
                    status = node['status']

                    if status == NodeStatus.RUNNING.value:
                        # If a spinner isn't already running for this node, start one.
                        if node_id not in self.active_spinners:
                            self.active_spinners.add(node_id)
                            self.update_spinner(tree_node, node)
                    else:
                        # For any non-running state, we are the source of truth.
                        # The spinner, if it exists, will see the state change and stop itself.
                        # We just set the final label.
                        if node.get('type') == 'block':
                            label = f"[b]{node_id}[/b]"
                        elif node.get('type') == 'info':
                            label = f"[cyan]i[/] {node_id}"
                        else:
                            icon = self.status_icons.get(status, " ")
                            label = f"{icon} {node_id}"
                        self.call_from_thread(tree_node.set_label, label)

                    # If the updated node is the one currently selected, refresh the action buttons
                    if node_id == self.selected_node_id:
                        if status == NodeStatus.FAILED.value and node.get('type') == 'playbook':
                            self.call_from_thread(self._set_widget_display, self.action_buttons, True)
                        else:
                            self.call_from_thread(self._set_widget_display, self.action_buttons, False)

                    if status == NodeStatus.AWAITING_CONFIRMATION.value:
                        if node_id not in self.approved_nodes and node_id not in self.pending_confirmation_nodes:
                            self.pending_confirmation_nodes.add(node_id)
                            message = f"Node [b]{node_id}[/b] is awaiting your confirmation."
                            if node.get('type') == 'checkpoint':
                                message = f"Checkpoint [b]{node_id}[/b] reached. Proceed?"
                            self.doubtful_node_queue.append((node_id, message))
                            nodes_need_approval = True

            if nodes_need_approval:
                self.call_from_thread(self._process_doubtful_queue)

        def on_tree_node_selected(self, event: Tree.NodeSelected) -> None:
            if self.stdout_watcher:
                self.stdout_watcher.cancel()
                self.stdout_watcher = None

            node_id = event.node.data
            self.selected_node_id = node_id
            node_data = self.node_data.get(node_id)

            if not node_data:
                self.action_buttons.display = False
                return

            self.details_table.clear()
            if not self.details_table.columns:
                self.details_table.add_column("Property", width=20)
                self.details_table.add_column("Value")

            def add_detail(key, value):
                self.details_table.add_row(key, value, height=None)

            add_detail("Node", node_data.get('id'))
            if node_data.get('type') == 'playbook':
                add_detail("Playbook", node_data.get('playbook', '-'))
                add_detail("Inventory", node_data.get('inventory', '-'))
                add_detail("Status", node_data.get('status', '-'))
                if node_data.get('started', False):
                    add_detail("Started", node_data.get('started', '-'))
                if node_data.get('ended', False):
                    add_detail("Ended", node_data.get('ended', '-'))
                if node_data.get('reference', False):
                    add_detail("Reference", node_data.get('reference', '-'))
                if node_data.get('description', False):
                    add_detail("Description", node_data.get('description', 'CCC'))
                if node_data.get('extravars', False):
                    add_detail("Variables", Pretty(node_data.get('extravars', {}), indent_guides=True, expand_all=False))
                self.show_stdout(node_id)
                if node_data['status'] == NodeStatus.RUNNING.value:
                    self.stdout_watcher = self.watch_stdout(node_id)
            elif node_data.get('type') == 'block':
                add_detail("Type", "Block")
                add_detail("Strategy", node_data.get('strategy', 'parallel'))
            elif node_data.get('type') == 'info':
                add_detail("Type", "Info")
                if node_data.get('description'):
                    add_detail("Description", node_data.get('description'))
                if node_data.get('reference'):
                    add_detail("Reference", node_data.get('reference'))
            elif node_data.get('type') == 'checkpoint':
                add_detail("Type", "Checkpoint")
                add_detail("Status", node_data.get('status', '-'))
                if node_data.get('description'):
                    add_detail("Description", node_data.get('description'))
                if node_data.get('reference'):
                    add_detail("Reference", node_data.get('reference'))

            if node_data.get('status') == NodeStatus.FAILED.value and node_data.get('type') == 'playbook':
                self.action_buttons.display = True
            else:
                self.action_buttons.display = False

        def on_button_pressed(self, event: Button.Pressed) -> None:
            """Called when a button is pressed."""
            if self.selected_node_id:
                if event.button.id == "relaunch_button":
                    self.api_client.restart_node(self.selected_node_id)
                    # Clear the log and start watching for new output
                    self.stdout_log.clear()
                    if self.stdout_watcher:
                        self.stdout_watcher.cancel()
                    self.stdout_watcher = self.watch_stdout(self.selected_node_id)
                elif event.button.id == "skip_button":
                    self.api_client.skip_node(self.selected_node_id)

            # Hide buttons after action
            self.action_buttons.display = False

        @work(exclusive=True, thread=True)
        def watch_stdout(self, node_id: str):
            last_content = self.api_client.get_node_stdout(node_id)
            if last_content is None:
                return

            while not self._shutdown_event.is_set():
                time.sleep(0.5)
                current_stdout = self.api_client.get_node_stdout(node_id)
                if current_stdout is None:
                    break
                if current_stdout != last_content:
                    new_content = current_stdout[len(last_content):]
                    text = Text.from_ansi(new_content)
                    self.call_from_thread(self.stdout_log.write, text)
                    last_content = current_stdout

                status_response = self.api_client.get_all_nodes()
                if status_response is None:
                    break
                node_status = next((n['status'] for n in status_response if n['id'] == node_id), None)
                if node_status != NodeStatus.RUNNING.value:
                    break

        @work(thread=True)
        def update_spinner(self, tree_node, node_data):
            """
            This worker is now self-terminating. It spins as long as the node's
            status is 'running' in the central self.node_data store.
            """
            spinner_cycle = itertools.cycle(self.spinner_icons)
            node_id = node_data['id']

            while self.node_data.get(node_id, {}).get('status') == NodeStatus.RUNNING.value and not self._shutdown_event.is_set():
                icon_char = next(spinner_cycle)
                icon = f"[yellow]{icon_char}[/yellow]"

                # Use the original node_data for static info like type and id
                if node_data.get('type') == 'block':
                    label = f"{icon} [b]{node_id}[/b]"
                else:
                    label = f"{icon} {node_id}"

                # Final check to prevent a race condition where the status changes
                # between the while-check and this set_label call.
                if self.node_data.get(node_id, {}).get('status') == NodeStatus.RUNNING.value:
                    self.call_from_thread(tree_node.set_label, label)

                time.sleep(0.1)

            # The loop has ended, meaning the node is no longer running.
            # The main update_node_statuses loop is now responsible for setting the
            # final label. This worker just needs to clean up its flag.
            if node_id in self.active_spinners:
                self.active_spinners.remove(node_id)

        @work(exclusive=True, thread=True)
        def show_stdout(self, node_id: str):
            """Reads and displays the entire stdout for a given node."""
            self.call_from_thread(self.stdout_log.clear)
            stdout = self.api_client.get_node_stdout(node_id)
            if stdout is not None:
                text = Text.from_ansi(stdout)
                self.call_from_thread(self.stdout_log.write, text)
